import { copy, normalizeArray, defineConfig, removeProperties, isValidObject } from '@agile-ts/utils';
import { LogCodeManager } from '../../logCodeManager.js';
import { GroupObserver } from './group.observer.js';
import { CollectionPersistent } from '../collection.persistent.js';
import { StateObserver } from '../../state/state.observer.js';
import { ComputedTracker } from '../../computed/computed.tracker.js';
import { EnhancedState } from '../../state/state.enhanced.js';

const _Group = class extends EnhancedState {
  constructor(collection, initialItems = [], config = {}) {
    super(collection.agileInstance(), initialItems, config);
    this._output = [];
    this.nextGroupOutput = [];
    this._preciseItemKeys = [];
    this.observers = {};
    this.notFoundItemKeys = [];
    this.loadedInitialValue = true;
    this.observers["value"] = new StateObserver(this, {
      key: config.key
    });
    this.observers["output"] = new GroupObserver(this, {
      key: config.key
    });
    this.collection = () => collection;
    this.addSideEffect(_Group.rebuildGroupSideEffectKey, (state, config2) => {
      var _a;
      this.rebuild(((_a = config2 == null ? void 0 : config2.any) == null ? void 0 : _a.trackedChanges) || [], config2);
    });
    this.rebuild();
  }
  get output() {
    ComputedTracker.tracked(this.observers["output"]);
    return copy(this._output);
  }
  set output(value) {
    LogCodeManager.log("1C:03:00", [this._key]);
  }
  has(itemKey) {
    return this.value.indexOf(itemKey) !== -1;
  }
  get size() {
    return this.value.length;
  }
  remove(itemKeys, config = {}) {
    const _itemKeys = normalizeArray(itemKeys);
    const notExistingItemKeysInCollection = [];
    const notExistingItemKeys = [];
    let newGroupValue = copy(this.nextStateValue);
    const updatedPreciseItemKeys = copy(this._preciseItemKeys);
    config = defineConfig(config, {
      softRebuild: true,
      any: {}
    });
    config.any["trackedChanges"] = [];
    _itemKeys.forEach((itemKey) => {
      const exists = newGroupValue.includes(itemKey);
      if (!exists) {
        notExistingItemKeys.push(itemKey);
        notExistingItemKeysInCollection.push(itemKey);
        return;
      }
      if (config.softRebuild) {
        const index = updatedPreciseItemKeys.findIndex((ik) => ik === itemKey);
        if (index !== -1) {
          updatedPreciseItemKeys.splice(index, 1);
          config.any["trackedChanges"].push({
            index,
            method: TrackedChangeMethod.REMOVE,
            key: itemKey
          });
        }
      }
      if (!this.collection().getItem(itemKey))
        notExistingItemKeysInCollection.push(itemKey);
      newGroupValue = newGroupValue.filter((key) => key !== itemKey);
    });
    if (notExistingItemKeys.length >= _itemKeys.length)
      return this;
    if (notExistingItemKeysInCollection.length >= _itemKeys.length)
      config.background = true;
    this.set(newGroupValue, removeProperties(config, ["softRebuild"]));
    return this;
  }
  add(itemKeys, config = {}) {
    const _itemKeys = normalizeArray(itemKeys);
    const notExistingItemKeysInCollection = [];
    const existingItemKeys = [];
    const newGroupValue = copy(this.nextStateValue);
    const updatedPreciseItemKeys = copy(this._preciseItemKeys);
    config = defineConfig(config, {
      method: "push",
      softRebuild: true,
      any: {}
    });
    config.any["trackedChanges"] = [];
    _itemKeys.forEach((itemKey) => {
      const exists = newGroupValue.includes(itemKey);
      if (!this.collection().getItem(itemKey))
        notExistingItemKeysInCollection.push(itemKey);
      if (exists) {
        existingItemKeys.push(itemKey);
        return;
      }
      if (config.softRebuild) {
        const index = config.method === "push" ? updatedPreciseItemKeys.length : 0;
        updatedPreciseItemKeys.push(itemKey);
        config.any["trackedChanges"].push({
          index,
          method: TrackedChangeMethod.ADD,
          key: itemKey
        });
      }
      newGroupValue[config.method || "push"](itemKey);
    });
    if (existingItemKeys.length >= _itemKeys.length)
      return this;
    if (notExistingItemKeysInCollection.concat(existingItemKeys).length >= _itemKeys.length)
      config.background = true;
    this.set(newGroupValue, removeProperties(config, ["method", "softRebuild"]));
    return this;
  }
  replace(oldItemKey, newItemKey, config = {}) {
    const newGroupValue = copy(this._value);
    newGroupValue.splice(newGroupValue.indexOf(oldItemKey), 1, newItemKey);
    this.set(newGroupValue, config);
    return this;
  }
  getItems() {
    return this.value.map((itemKey) => this.collection().getItem(itemKey)).filter((item) => item !== void 0);
  }
  persist(keyOrConfig = {}, config = {}) {
    let _config;
    let key;
    if (isValidObject(keyOrConfig)) {
      _config = keyOrConfig;
      key = this._key;
    } else {
      _config = config || {};
      key = keyOrConfig;
    }
    _config = defineConfig(_config, {
      loadValue: true,
      followCollectionPersistKeyPattern: true,
      storageKeys: [],
      defaultStorageKey: null
    });
    if (_config.followCollectionPersistKeyPattern) {
      key = CollectionPersistent.getGroupStorageKey(key || this._key, this.collection()._key);
    }
    super.persist(key, {
      loadValue: _config.loadValue,
      storageKeys: _config.storageKeys,
      defaultStorageKey: _config.defaultStorageKey
    });
    return this;
  }
  rebuild(trackedChanges = [], config = {}) {
    if (!this.collection().isInstantiated)
      return this;
    const notFoundItemKeys = [];
    if (trackedChanges.length > 0) {
      trackedChanges.forEach((change) => {
        const item = this.collection().getItem(change.key);
        switch (change.method) {
          case TrackedChangeMethod.ADD:
            if (item != null) {
              this._preciseItemKeys.splice(change.index, 0, change.key);
              this.nextGroupOutput.splice(change.index, 0, copy(item._value));
            } else {
              notFoundItemKeys.push(change.key);
            }
            break;
          case TrackedChangeMethod.UPDATE:
            if (item != null) {
              this.nextGroupOutput[change.index] = copy(item._value);
            } else {
              notFoundItemKeys.push(change.key);
            }
            break;
          case TrackedChangeMethod.REMOVE:
            this._preciseItemKeys.splice(change.index, 1);
            this.nextGroupOutput.splice(change.index, 1);
            break;
        }
      });
      this.observers["output"].ingest(config);
    } else {
      const groupItemValues = [];
      this._preciseItemKeys = [];
      this._value.forEach((itemKey) => {
        const item = this.collection().getItem(itemKey);
        if (item != null) {
          groupItemValues.push(item._value);
          this._preciseItemKeys.push(itemKey);
        } else
          notFoundItemKeys.push(itemKey);
      });
      this.observers["output"].ingestOutput(groupItemValues, config);
    }
    if (notFoundItemKeys.length > 0 && this.loadedInitialValue) {
      LogCodeManager.log("1C:02:00", [this.collection()._key, this._key], notFoundItemKeys);
    }
    this.notFoundItemKeys = notFoundItemKeys;
    return this;
  }
};
let Group = _Group;
Group.rebuildGroupSideEffectKey = "rebuildGroup";
var TrackedChangeMethod;
(function(TrackedChangeMethod2) {
  TrackedChangeMethod2[TrackedChangeMethod2["ADD"] = 0] = "ADD";
  TrackedChangeMethod2[TrackedChangeMethod2["REMOVE"] = 1] = "REMOVE";
  TrackedChangeMethod2[TrackedChangeMethod2["UPDATE"] = 2] = "UPDATE";
})(TrackedChangeMethod || (TrackedChangeMethod = {}));

export { Group, TrackedChangeMethod };
