'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var utils = require('@agile-ts/utils');

// The Log Code Manager keeps track
// and manages all important Logs of AgileTs.
//
// How does the identification of Log Messages work?
// Let's take a look at this example:
// 00:00:00
//
// |00|:00:00 first digits are based on the Agile Class
// 00 = General
// 10 = Agile
// 11 = Storage
// ...
//
// ---
// 00:|00|:00 second digits are based on the Log Type
const logCodeTypes = {
    '00': 'success',
    '01': 'info',
    '02': 'warn',
    '03': 'error',
};
//
// ---
// 00:00:|00| third digits are based on the Log Message (ascending counted)
let allowLogging = true;
const niceLogCodeMessages = {
    // Agile
    '10:00:00': 'Created new AgileInstance.',
    '10:02:00': 'Be careful when binding multiple Agile Instances globally in one application!',
    // Storages
    '11:02:00': "The 'Local Storage' is not available in your current environment. " +
        "To use the '.persist()' functionality, please provide a custom Storage!",
    '11:02:01': 'The first allocated Storage for AgileTs must be set as the default Storage!',
    '11:03:00': "Storage with the key/name '${0}' already exists!",
    '11:03:01': "Couldn't find Storage '${0}'. " +
        "The Storage with the key/name '${0}' doesn't exists!",
    '11:03:02': "Storage with the key/name '${0}' isn't ready yet!",
    '11:03:03': 'No Storage found to get a value from! Please specify at least one Storage.',
    '11:03:04': 'No Storage found to store a value in! Please specify at least one Storage.',
    '11:03:05': 'No Storage found to remove a value from! Please specify at least one Storage.',
    // Persistent
    '12:03:00': 'No valid persist Key found! Provide a valid key or assign one to the parent instance.',
    '12:03:01': 'No valid persist Storage Key found! Please specify at least one Storage Key to use the persist functionality.',
    '12:03:02': "Couldn't validate Persistent '${0}'." +
        "The Storage with the key/name '${1}' doesn't exists!`",
    // Storage
    '13:00:00': "Registered new Storage '${0}'.",
    '13:01:00': "GET value at key '${1}' from Storage '${0}'.",
    '13:01:01': "SET value at key '${1}' in Storage '${0}'.",
    '13:01:02': "REMOVE value at key '${1}' from Storage '${0}'.",
    '13:02:00': 'Using normalGet() in a async-based Storage might result in an unexpected return value. ' +
        'Instead of a resolved value a Promise is returned!',
    '13:03:00': "Invalid Storage '${0}()' method provided!",
    // State
    '14:03:01': "'${1}' is a not supported type! Supported types are: String, Boolean, Array, Object, Number.",
    '14:03:02': "The 'patch()' method works only in object based States!",
    '14:03:03': 'Only one Interval can be active at once!',
    '14:03:04': "Failed to invert value of the type '${0}'!",
    // SubController
    '15:01:00': "Unregistered 'Callback' based Subscription.",
    '15:01:01': "Unregistered 'Component' based Subscription.",
    '15:01:02': "Registered 'Component' based Subscription.",
    '15:01:03': "Registered 'Callback' based Subscription.",
    // Runtime
    '16:01:00': "Created Job '${0}'",
    '16:01:01': "Completed Job '${0}'",
    '16:01:02': 'Updated/Rerendered Subscriptions',
    '16:02:00': "SubscriptionContainer/Component '${0}' isn't ready to rerender!",
    '16:02:01': 'Job with not ready SubscriptionContainer/Component was removed from the runtime ' +
        'after ${0} tries to avoid a Job overflow.',
    // Observer
    '17:03:00': "The 'perform()' method isn't set in Observer but need to be set! Observer is no stand alone class.",
    // Integrations
    '18:00:00': "Integrated '${0}' into AgileTs '${1}'",
    '18:02:00': "Can't call the 'update()' method on a not ready Integration '${0}'!",
    '18:03:00': "Failed to integrate Framework '${0}' into AgileTs '${1}'!",
    // Computed
    // Collection Persistent
    '1A:02:00': 'Failed to build unique Item StorageKey!',
    '1A:02:01': 'Failed to build unique Group StorageKey!',
    // Collection
    '1B:02:00': "We recommend using 'createGroup()' " +
        "instead of 'Group()' outside the Collection configuration object.",
    '1B:02:01': "We recommend using 'createSelector()' " +
        "instead of 'Selector()' outside the Collection configuration object.",
    '1B:02:02': 'By overwriting the whole Item ' +
        "you have to pass the correct itemKey into the 'changes object!'",
    '1B:02:03': "We recommend using 'Group()' instead of 'createGroup()' " +
        'inside the Collection configuration object.',
    '1B:02:04': "We recommend using 'Selector()' instead of 'createSelector()' " +
        'inside the Collection configuration object.',
    '1B:02:05': "Collection '${0}' Item Data has to contain a primaryKey property called '${1}'!",
    '1B:03:00': "Couldn't update Item with the key/name '${0}' " +
        "because it doesn't exist in Collection '${1}'",
    '1B:03:01': "Valid object required to update Item value '${0}' in Collection '${1}'!",
    '1B:03:02': "Group with the key/name '${0}' already exists!",
    '1B:03:03': "Selector with the key/name '${0}' already exists!",
    '1B:03:04': "Couldn't update ItemKey from '${0}' to '${1}' " +
        "because an Item with the key/name '${1}' already exists in the Collection '${2}'!",
    '1B:03:05': "Item Data of Collection '${0}' has to be a valid object!",
    '1B:03:06': "Item tried to add to the Collection '${0}' belongs to another Collection '${1}'!",
    // Group
    '1C:02:00': "Couldn't find some Items in the Collection '${0}' " +
        "during the rebuild of the Group '${1}' output.",
    '1C:03:00': "The 'output' property of the Group '${0}' is a automatically generated readonly property " +
        'that can only be mutated by the Group itself!',
    '1C:03:01': "The 'item' property of the Group '${0}' is a automatically generated readonly property " +
        'that can only be mutated by the Group itself!',
    // Utils
    '20:03:00': 'Failed to get Agile Instance from',
    '20:03:01': "Failed to create global Instance at '${0}'",
    '20:03:02': "Required module '${0}' couldn't be retrieved!",
    // General
    '00:03:00': "The '${0}()' method isn't set in ${1} but need to be set!" +
        ' ${1} is no stand alone class.',
    '00:03:01': "'${0}' has to be of the type ${1}!",
};
// Note: Not outsource the 'production' env check,
// because then webpack can't treeshake based on the current env
const logCodeMessages = process.env.NODE_ENV !== 'production' ? niceLogCodeMessages : {};
/**
 * Specifies whether the LogCodeManager is allowed to print any logs.
 *
 * @internal
 * @param logging - Whether the LogCodeManager is allowed to print any logs.
 */
function setAllowLogging(logging) {
    allowLogging = logging;
}
/**
 * Returns the log message according to the specified log code.
 *
 * @internal
 * @param logCode - Log code of the message to be returned.
 * @param replacers - Instances that replace these '${x}' placeholders based on the index
 * For example: 'replacers[0]' replaces '${0}', 'replacers[1]' replaces '${1}', ..
 */
function getLog(logCode, replacers = []) {
    let result = logCodeMessages[logCode];
    if (result == null)
        return logCode;
    // Replace '${x}' with the specified replacer instances
    for (let i = 0; i < replacers.length; i++) {
        result = result.replace('${' + i + '}', replacers[i]);
    }
    return result;
}
/**
 * Logs the log message according to the specified log code
 * with the Agile Logger if installed or the normal console.
 *
 * @internal
 * @param logCode - Log code of the message to be returned.
 * @param replacers - Instances that replace these '${x}' placeholders based on the index
 * For example: 'replacers[0]' replaces '${0}', 'replacers[1]' replaces '${1}', ..
 * @param data - Data to be attached to the end of the log message.
 */
function log(logCode, replacers = [], ...data) {
    const logger = LogCodeManager.getLogger();
    if ((logger != null && !logger.isActive) || !allowLogging)
        return;
    const logType = logCodeTypes[logCode.substr(3, 2)];
    if (typeof logType !== 'string')
        return;
    // Handle logging without Logger
    if (logger == null) {
        if (logType === 'error' || logType === 'warn')
            console[logType](`Agile: ${getLog(logCode, replacers)}`);
        return;
    }
    // Handle logging with Logger
    logger[logType](getLog(logCode, replacers), ...data);
}
/**
 * Logs the log message according to the specified log code
 * with the Agile Logger if installed and the provided tags are active.
 *
 * @internal
 * @param tags - Tags to be active to log the logCode.
 * @param logCode - Log code of the message to be returned.
 * @param replacers - Instances that replace these '${x}' placeholders based on the index
 * For example: 'replacers[0]' replaces '${0}', 'replacers[1]' replaces '${1}', ..
 * @param data - Data to be attached to the end of the log message.
 */
function logIfTags(tags, logCode, replacers = [], ...data) {
    const logger = LogCodeManager.getLogger();
    if ((logger != null && !logger.isActive) || !allowLogging)
        return;
    const logType = logCodeTypes[logCode.substr(3, 2)];
    if (typeof logType !== 'string')
        return;
    // Handle logging without Logger
    if (logger == null) {
        // Log nothing since if a log has a tag it is probably not so important
        return;
    }
    // Handle logging with Logger
    logger.if.tag(tags)[logType](getLog(logCode, replacers), ...data);
}
/**
 * Creates an extension of the specified LogCodeManager
 * and assigns the provided additional log messages to it.
 *
 * @param additionalLogs - Log messages to be added to the LogCodeManager.
 * @param logCodeManager - LogCodeManager to create an extension from.
 */
function assignAdditionalLogs(additionalLogs, logCodeManager) {
    const copiedLogCodeManager = utils.copy(logCodeManager);
    copiedLogCodeManager.logCodeMessages = {
        ...copiedLogCodeManager.logCodeMessages,
        ...additionalLogs,
    };
    return copiedLogCodeManager;
}
let tempLogCodeManager;
if (process.env.NODE_ENV !== 'production') {
    let loggerPackage = null;
    try {
        loggerPackage = require('@agile-ts/logger');
    }
    catch (e) {
        // empty catch block
    }
    tempLogCodeManager = {
        getLog,
        log,
        logCodeLogTypes: logCodeTypes,
        logCodeMessages: logCodeMessages,
        getLogger: () => {
            // https://github.com/agile-ts/agile/issues/203
            if (loggerPackage != null)
                return loggerPackage.sharedLogger;
            return null;
        },
        logIfTags,
        setAllowLogging,
    };
}
else {
    tempLogCodeManager = {
        // Log only logCode
        getLog: (logCode, replacers) => logCode,
        log,
        logCodeLogTypes: logCodeTypes,
        logCodeMessages: {},
        getLogger: () => {
            return null;
        },
        logIfTags: (tags, logCode, replacers) => {
            /* empty because logs with tags can't be that important */
        },
        setAllowLogging,
    };
}
/**
 * The Log Code Manager keeps track
 * and manages all important Logs for the '@agile-ts/core' package.
 *
 * @internal
 */
const LogCodeManager = tempLogCodeManager;

class SubscriptionContainer {
    /**
     * Key/Name identifier of the Subscription Container.
     */
    key;
    /**
     * Whether the Subscription Container
     * and the UI-Component it represents are ready.
     *
     * When both instances are ready,
     * the Subscription Container is allowed
     * to trigger re-renders on the UI-Component.
     */
    ready = false;
    /**
     * Unique identifier of the UI-Component
     * the Subscription Container represents.
     */
    componentId;
    /**
     * Observers that are subscribed to the Subscription Container.
     *
     * The subscribed Observers use the Subscription Container
     * as an interface to the UI-Component it represents.
     *
     * Through the Subscription Container, the Observers can easily trigger re-renders
     * on the UI-Component, for example, when their value updates.
     *
     * [Learn more..](https://agile-ts.org/docs/core/integration#-subscriptions)
     */
    subscribers;
    /**
     * Temporary stores the subscribed Observers,
     * that were updated by the runtime
     * and are currently running through
     * the update (rerender) Subscription Container (UI-Component) process.
     */
    updatedSubscribers = new Set();
    /**
     * Whether the Subscription Container is object based.
     *
     * A Subscription Container is object based when the subscribed Observers
     * have been provided in an Observer keymap object
     * ```
     * {
     *   state1: Observer,
     *   state2: Observer
     * }
     * ```
     * Thus each Observer has its 'external' unique key stored in the `subscribersWeakMap`.
     *
     * Often Component based Subscriptions are object based,
     * because each Observer requires in such Subscription a unique identifier.
     * Mainly to be properly represented in the `updatedData` object
     * sent to the Integration's `updateMethod()` method
     * when the Subscription Container updates (re-renders the UI-Component).
     */
    isObjectBased = false;
    /**
     * Weak map for storing 'external' key identifiers for subscribed Observers.
     *
     * Why is the key not applied directly to the Observer?
     *
     * Because the key defined here should be only valid
     * for the scope of the Subscription Container.
     *
     * https://stackoverflow.com/questions/29413222/what-are-the-actual-uses-of-es6-weakmap
     */
    subscriberKeysWeakMap;
    /**
     * Weak Map for storing selector functions for subscribed Observers.
     *
     * A selector function allows the partial subscription to an Observer value.
     * Only when the selected Observer value part changes,
     * the Subscription Container is updated (-> re-renders the UI-Component).
     *
     * Why are the selector functions not applied directly to the Observer?
     *
     * Because the selector function defined here should be only valid
     * for the scope of the Subscription Container.
     *
     * https://stackoverflow.com/questions/29413222/what-are-the-actual-uses-of-es6-weakmap
     */
    selectorsWeakMap;
    /**
     * A Subscription Container represents a UI-Component in AgileTs
     * that can be subscribed by multiple Observer Instances.
     *
     * The subscribed Observers can use the Subscription Container as an interface
     * to the UI-Component it represents.
     * For example, to trigger re-renders on the UI-Component,
     * when their value has changed.
     *
     * @internal
     * @param subs - Observers to be initial subscribed to the Subscription Container.
     * @param config - Configuration object
     */
    constructor(subs, config = {}) {
        config = utils.defineConfig(config, {
            key: utils.generateId(),
        });
        this.subscribers = new Set();
        this.key = config.key;
        this.componentId = config?.componentId;
        this.subscriberKeysWeakMap = new WeakMap();
        this.selectorsWeakMap = new WeakMap();
        this.isObjectBased = !Array.isArray(subs);
        // Assign initial Observers to the Subscription Container
        for (const key in subs) {
            this.addSubscription(subs[key], {
                proxyPaths: config.proxyWeakMap?.get(subs[key])?.paths,
                selectorMethods: config.selectorWeakMap?.get(subs[key])?.methods,
                key: this.isObjectBased ? key : undefined,
            });
        }
    }
    /**
     * Subscribes the specified Observer to the Subscription Container.
     *
     * @internal
     * @param sub - Observer to be subscribed to the Subscription Container
     * @param config - Configuration object
     */
    addSubscription(sub, config = {}) {
        const toAddSelectorMethods = config.selectorMethods ?? [];
        const proxyPaths = config.proxyPaths ?? [];
        // Create additional selector methods based on the specified proxy paths
        for (const path of proxyPaths) {
            toAddSelectorMethods.push((value) => {
                let _value = value;
                for (const branch of path) {
                    if (!utils.isValidObject(_value, true))
                        break;
                    _value = _value[branch];
                }
                return _value;
            });
        }
        // Assign defined/created selector methods to the 'selectorsWeakMap'
        // (Not to the Observer itself, since the selector methods specified here
        // only count for the scope of the Subscription Container)
        const existingSelectorMethods = this.selectorsWeakMap.get(sub)?.methods ?? [];
        const newSelectorMethods = existingSelectorMethods.concat(toAddSelectorMethods);
        if (newSelectorMethods.length > 0)
            this.selectorsWeakMap.set(sub, { methods: newSelectorMethods });
        // Assign specified key to the 'subscriberKeysWeakMap'
        // (Not to the Observer itself, since the key specified here
        // only counts for the scope of the Subscription Container)
        if (config.key != null)
            this.subscriberKeysWeakMap.set(sub, config.key);
        // Add Observer to subscribers
        this.subscribers.add(sub);
        // Add Subscription Container to Observer
        // so that the Observer can cause updates on it
        // (trigger re-render on the UI-Component it represents).
        sub.subscribedTo.add(this);
    }
    /**
     * Unsubscribes the specified Observer from the Subscription Container.
     *
     * @internal
     * @param sub - Observer to be unsubscribed from the Subscription Container.
     */
    removeSubscription(sub) {
        if (this.subscribers.has(sub)) {
            this.selectorsWeakMap.delete(sub);
            this.subscriberKeysWeakMap.delete(sub);
            this.subscribers.delete(sub);
            sub.subscribedTo.delete(this);
        }
    }
}

class CallbackSubscriptionContainer extends SubscriptionContainer {
    /**
     * Callback function to trigger a re-render
     * on the UI-Component which is represented by the Subscription Container.
     */
    callback;
    /**
     * A Callback Subscription Container represents a UI-Component in AgileTs
     * and triggers re-renders on the UI-Component via the specified callback function.
     *
     * The Callback Subscription Container doesn't keep track of the Component itself.
     * It only knows how to trigger re-renders on it by calling the callback function.
     *
     * [Learn more..](https://agile-ts.org/docs/core/integration#callback-based)
     *
     * @internal
     * @param callback - Callback function to cause a rerender on the Component
     * to be represented by the Subscription Container.
     * @param subs - Observers to be initial subscribed to the Subscription Container.
     * @param config - Configuration object
     */
    constructor(callback, subs, config = {}) {
        super(subs, config);
        this.callback = callback;
    }
}

class ComponentSubscriptionContainer extends SubscriptionContainer {
    /**
     * UI-Component which is represented by the Subscription Container
     * and mutated via the Integration's `updateMethod()` method
     * to cause re-renders on it.
     */
    component;
    /**
     * A Component Subscription Container represents a UI-Component in AgileTs
     * and triggers re-renders on the UI-Component by muting the specified Component Instance
     * via the Integration's `updateMethod()` method.
     * For example by updating a local State Management property of the Component
     * (like in React Class Components the `this.state` property).
     *
     * The Component Subscription Container keeps track of the Component itself,
     * to mutate it appropriately so that re-renders can be triggered on it.
     *
     * For this to work well, a Component Subscription Container is often object based.
     * Meaning that each Observer was provided in an object keymap
     * with a unique key identifier.
     * ```
     * // Object based (guaranteed unique key identifier)
     * {
     *   state1: Observer,
     *   state2: Observer
     * }
     *
     * // Array based (no guaranteed unique key identifier)
     * [Observer, Observer]
     * ```
     * Thus the Integration's 'updateMethod()' method can be called
     * with a complete object of updated Observer values.
     * ```
     * updateMethod: (componentInstance, updatedData) => {
     *   console.log(componentInstance); // Returns 'this.component'
     *   console.log(updatedData); // Returns updated Observer values keymap (see below)
     *   // {
     *   //   state1: Observer.value,
     *   //   state2: Observer.value,
     *   // }
     *   }
     * ```
     *
     * [Learn more..](https://agile-ts.org/docs/core/integration#component-based)
     *
     * @internal
     * @param component - UI-Component to be represented by the Subscription Container
     * and mutated via the Integration's 'updateMethod()' method to trigger re-renders on it.
     * @param subs - Observers to be initial subscribed to the Subscription Container.
     * @param config - Configuration object
     */
    constructor(component, subs, config = {}) {
        super(subs, config);
        this.component = component;
    }
}

class SubController {
    // Agile Instance the SubController belongs to
    agileInstance;
    // Keeps track of all registered Component based Subscriptions
    componentSubs = new Set();
    // Keeps track of all registered Callback based Subscriptions
    callbackSubs = new Set();
    // Keeps track of all mounted Components (only if agileInstance.config.mount = true)
    mountedComponents = new Set();
    /**
     * The Subscription Controller manages and simplifies the subscription to UI-Components.
     *
     * Thus it creates Subscription Containers (Interfaces to UI-Components)
     * and assigns them to the specified Observers.
     * These Observers can then easily trigger re-renders on UI-Components
     * via the created Subscription Containers.
     *
     * @internal
     * @param agileInstance - Instance of Agile the Subscription Controller belongs to.
     */
    constructor(agileInstance) {
        this.agileInstance = () => agileInstance;
    }
    subscribe(integrationInstance, subs, config = {}) {
        config = utils.defineConfig(config, {
            waitForMount: this.agileInstance().config.waitForMount,
        });
        // Create Subscription Container based on specified 'integrationInstance'
        const subscriptionContainer = utils.isFunction(integrationInstance)
            ? this.createCallbackSubscriptionContainer(integrationInstance, subs, config)
            : this.createComponentSubscriptionContainer(integrationInstance, subs, config);
        // Return object based Subscription Container and an Observer value keymap
        if (subscriptionContainer.isObjectBased && !Array.isArray(subs)) {
            const props = {};
            for (const key in subs)
                if (subs[key].value)
                    props[key] = subs[key].value;
            return { subscriptionContainer, props };
        }
        // Return array based Subscription Container
        return subscriptionContainer;
    }
    /**
     * Removes the Subscription Container extracted from the specified 'subscriptionInstance'
     * from all Observers that were subscribed to it.
     *
     * We should always unregister a Subscription Container when it is no longer in use.
     * For example, when the UI-Component it represents has been unmounted.
     *
     * @public
     * @param subscriptionInstance - UI-Component that contains an instance of a Subscription Container
     * or a Subscription Container to be unsubscribed/unregistered.
     */
    unsubscribe(subscriptionInstance) {
        // Helper function to remove Subscription Container from Observer
        const unsub = (subscriptionContainer) => {
            subscriptionContainer.ready = false;
            subscriptionContainer.subscribers.forEach((observer) => {
                subscriptionContainer.removeSubscription(observer);
            });
        };
        // Unsubscribe Callback based Subscription Container
        if (subscriptionInstance instanceof CallbackSubscriptionContainer) {
            unsub(subscriptionInstance);
            this.callbackSubs.delete(subscriptionInstance);
            LogCodeManager.logIfTags(['subscription'], '15:01:00', [], subscriptionInstance);
            return;
        }
        // Unsubscribe Component based Subscription Container
        if (subscriptionInstance instanceof ComponentSubscriptionContainer) {
            unsub(subscriptionInstance);
            this.componentSubs.delete(subscriptionInstance);
            LogCodeManager.logIfTags(['subscription'], '15:01:01', [], subscriptionInstance);
            return;
        }
        // Unsubscribe Component based Subscription Container
        // extracted from the 'componentSubscriptionContainers' property
        if (subscriptionInstance['componentSubscriptionContainers'] !== null &&
            Array.isArray(subscriptionInstance.componentSubscriptionContainers)) {
            subscriptionInstance.componentSubscriptionContainers.forEach((subContainer) => {
                unsub(subContainer);
                this.componentSubs.delete(subContainer);
                LogCodeManager.logIfTags(['subscription'], '15:01:01', [], subscriptionInstance);
            });
            return;
        }
    }
    /**
     * Returns a newly created Component based Subscription Container.
     *
     * @internal
     * @param componentInstance - UI-Component to be represented by the Subscription Container
     * and mutated via the Integration's 'updateMethod()' method to trigger re-renders on it.
     * @param subs - Observers to be initial subscribed to the Subscription Container.
     * @param config - Configuration object.
     */
    createComponentSubscriptionContainer(componentInstance, subs, config = {}) {
        const componentSubscriptionContainer = new ComponentSubscriptionContainer(componentInstance, subs, utils.removeProperties(config, ['waitForMount']));
        this.componentSubs.add(componentSubscriptionContainer);
        // Define ready state of Subscription Container
        if (config.waitForMount) {
            if (this.mountedComponents.has(componentInstance))
                componentSubscriptionContainer.ready = true;
        }
        else
            componentSubscriptionContainer.ready = true;
        // Add Subscription Container to the UI-Component it represents.
        // (For example, useful to unsubscribe the Subscription Container via the Component Instance)
        if (componentInstance['componentSubscriptionContainers'] != null &&
            Array.isArray(componentInstance.componentSubscriptionContainers))
            componentInstance.componentSubscriptionContainers.push(componentSubscriptionContainer);
        else
            componentInstance['componentSubscriptionContainers'] = [
                componentSubscriptionContainer,
            ];
        LogCodeManager.logIfTags(['subscription'], '15:01:02', [], componentSubscriptionContainer);
        return componentSubscriptionContainer;
    }
    /**
     * Returns a newly created Callback based Subscription Container.
     *
     * @internal
     * @param callbackFunction - Callback function to cause a rerender on the Component
     * to be represented by the Subscription Container.
     * @param subs - Observers to be initial subscribed to the Subscription Container.
     * @param config - Configuration object
     */
    createCallbackSubscriptionContainer(callbackFunction, subs, config = {}) {
        const callbackSubscriptionContainer = new CallbackSubscriptionContainer(callbackFunction, subs, utils.removeProperties(config, ['waitForMount']));
        this.callbackSubs.add(callbackSubscriptionContainer);
        callbackSubscriptionContainer.ready = true;
        LogCodeManager.logIfTags(['subscription'], '15:01:03', [], callbackSubscriptionContainer);
        return callbackSubscriptionContainer;
    }
    /**
     * Notifies the Subscription Containers representing the specified UI-Component (`componentInstance`)
     * that the UI-Component they represent has been mounted.
     *
     * @public
     * @param componentInstance - Component Instance containing Subscription Containers to be mounted.
     */
    mount(componentInstance) {
        if (componentInstance['componentSubscriptionContainers'] != null &&
            Array.isArray(componentInstance.componentSubscriptionContainers))
            componentInstance.componentSubscriptionContainers.map((c) => (c.ready = true));
        this.mountedComponents.add(componentInstance);
    }
    /**
     * Notifies the Subscription Containers representing the specified UI-Component (`componentInstance`)
     * that the UI-Component they represent has been unmounted.
     *
     * @public
     * @param componentInstance - Component Instance containing Subscription Containers to be unmounted
     */
    unmount(componentInstance) {
        if (componentInstance['componentSubscriptionContainers'] != null &&
            Array.isArray(componentInstance.componentSubscriptionContainers))
            componentInstance.componentSubscriptionContainers.map((c) => (c.ready = false));
        this.mountedComponents.delete(componentInstance);
    }
}

class Runtime {
    // Agile Instance the Runtime belongs to
    agileInstance;
    // Job that is currently being performed
    currentJob = null;
    // Jobs to be performed
    jobQueue = [];
    // Jobs that were performed and are ready to re-render
    jobsToRerender = [];
    // Jobs that were performed and couldn't be re-rendered yet.
    // That is the case when at least one Subscription Container (UI-Component) in the Job
    // wasn't ready to update (re-render).
    notReadyJobsToRerender = new Set();
    // Whether the `jobQueue` is currently being actively processed
    isPerformingJobs = false;
    // Current 'bucket' timeout 'scheduled' for updating the Subscribers (UI-Components)
    bucketTimeout = null;
    /**
     * The Runtime queues and executes incoming Observer-based Jobs
     * to prevent [race conditions](https://en.wikipedia.org/wiki/Race_condition#:~:text=A%20race%20condition%20or%20race,the%20possible%20behaviors%20is%20undesirable.)
     * and optimized the re-rendering of the Observer's subscribed UI-Components.
     *
     * Each queued Job is executed when it is its turn
     * by calling the Job Observer's `perform()` method.
     *
     * After successful execution, the Job is added to a re-render queue,
     * which is first put into the browser's 'Bucket' and started to work off
     * when resources are left.
     *
     * The re-render queue is designed for optimizing the render count
     * by batching multiple re-render Jobs of the same UI-Component
     * and ignoring re-render requests for unmounted UI-Components.
     *
     * @internal
     * @param agileInstance - Instance of Agile the Runtime belongs to.
     */
    constructor(agileInstance) {
        this.agileInstance = () => agileInstance;
    }
    /**
     * Adds the specified Observer-based Job to the internal Job queue,
     * where it is executed when it is its turn.
     *
     * After successful execution, the Job is assigned to the re-render queue,
     * where all the Observer's subscribed Subscription Containers (UI-Components)
     * are updated (re-rendered).
     *
     * @public
     * @param job - Job to be added to the Job queue.
     * @param config - Configuration object
     */
    ingest(job, config = {}) {
        config = utils.defineConfig(config, {
            perform: !this.isPerformingJobs,
        });
        // Add specified Job to the queue
        this.jobQueue.push(job);
        LogCodeManager.logIfTags(['runtime'], '16:01:00', [job._key], job);
        // Run first Job from the queue
        if (config.perform) {
            const performJob = this.jobQueue.shift();
            if (performJob)
                this.perform(performJob);
        }
    }
    /**
     * Performs the specified Job
     * and assigns it to the re-render queue if necessary.
     *
     * After the execution of the provided Job, it is checked whether
     * there are still Jobs left in the Job queue.
     * - If so, the next Job in the `jobQueue` is performed.
     * - If not, the `jobsToRerender` queue is started to work off.
     *
     * @internal
     * @param job - Job to be performed.
     */
    perform(job) {
        this.isPerformingJobs = true;
        this.currentJob = job;
        // Perform Job
        job.observer.perform(job);
        job.performed = true;
        // Ingest dependents of the Observer into runtime,
        // since they depend on the Observer and therefore have properly changed too
        job.observer.dependents.forEach((observer) => observer.ingest());
        // Add Job to rerender queue and reset current Job property
        if (job.rerender)
            this.jobsToRerender.push(job);
        this.currentJob = null;
        LogCodeManager.logIfTags(['runtime'], '16:01:01', [job._key], job);
        // Perform Jobs as long as Jobs are left in the queue.
        // If no Job is left start updating (re-rendering) Subscription Container (UI-Components)
        // of the Job based on the 'jobsToRerender' queue.
        if (this.jobQueue.length > 0) {
            const performJob = this.jobQueue.shift();
            if (performJob)
                this.perform(performJob);
        }
        else {
            this.isPerformingJobs = false;
            if (this.jobsToRerender.length > 0) {
                if (this.agileInstance().config.bucket) {
                    // Check if an bucket timeout is active, if so don't call a new one,
                    // since if the active timeout is called it will also proceed Jobs
                    // that were not added before the call
                    if (this.bucketTimeout == null) {
                        // https://stackoverflow.com/questions/9083594/call-settimeout-without-delay
                        this.bucketTimeout = setTimeout(() => {
                            this.bucketTimeout = null;
                            this.updateSubscribers();
                        });
                    }
                }
                else
                    this.updateSubscribers();
            }
        }
    }
    /**
     * Processes the `jobsToRerender` queue by updating (causing a re-render on)
     * the subscribed Subscription Containers (UI-Components) of each Job Observer.
     *
     * It returns a boolean indicating whether
     * any Subscription Container (UI-Component) was updated (re-rendered) or not.
     *
     * @internal
     */
    updateSubscribers() {
        // Build final 'jobsToRerender' array
        // based on the new 'jobsToRerender' array and the 'notReadyJobsToRerender' array
        const jobsToRerender = this.jobsToRerender.concat(Array.from(this.notReadyJobsToRerender));
        this.notReadyJobsToRerender = new Set();
        this.jobsToRerender = [];
        if (!this.agileInstance().hasIntegration() || jobsToRerender.length <= 0)
            return false;
        // Extract the Subscription Container to be re-rendered from the Jobs
        const subscriptionContainerToUpdate = this.extractToUpdateSubscriptionContainer(jobsToRerender);
        if (subscriptionContainerToUpdate.length <= 0)
            return false;
        // Update Subscription Container (trigger re-render on the UI-Component they represent)
        this.updateSubscriptionContainer(subscriptionContainerToUpdate);
        return true;
    }
    /**
     * Extracts the Subscription Containers (UI-Components)
     * to be updated (re-rendered) from the specified Runtime Jobs.
     *
     * @internal
     * @param jobs - Jobs from which to extract the Subscription Containers to be updated.
     */
    extractToUpdateSubscriptionContainer(jobs) {
        // https://medium.com/@bretcameron/how-to-make-your-code-faster-using-javascript-sets-b432457a4a77
        const subscriptionsToUpdate = new Set();
        // Using for loop for performance optimization
        // https://stackoverflow.com/questions/43821759/why-array-foreach-is-slower-than-for-loop-in-javascript
        for (let i = 0; i < jobs.length; i++) {
            const job = jobs[i];
            job.subscriptionContainersToUpdate.forEach((subscriptionContainer) => {
                let updateSubscriptionContainer = true;
                // Handle not ready Subscription Container
                if (!subscriptionContainer.ready) {
                    if (!job.config.maxTriesToUpdate ||
                        job.timesTriedToUpdateCount < job.config.maxTriesToUpdate) {
                        job.timesTriedToUpdateCount++;
                        this.notReadyJobsToRerender.add(job);
                        LogCodeManager.log('16:02:00', [subscriptionContainer.key], subscriptionContainer);
                    }
                    else {
                        LogCodeManager.log('16:02:01', [job.config.maxTriesToUpdate], subscriptionContainer);
                    }
                    return;
                }
                // TODO has to be overthought because when it is a Component based Subscription
                //  the rerender is triggered via merging the changed properties into the Component.
                //  Although the 'componentId' might be equal, it doesn't mean
                //  that the changed properties are equal! (-> changed properties might get missing)
                // Check if Subscription Container with same 'componentId'
                // is already in the 'subscriptionToUpdate' queue (rerender optimisation)
                // updateSubscriptionContainer =
                //   updateSubscriptionContainer &&
                //   Array.from(subscriptionsToUpdate).findIndex(
                //     (sc) => sc.componentId === subscriptionContainer.componentId
                //   ) === -1;
                // Check whether a selected part of the Observer value has changed
                updateSubscriptionContainer =
                    updateSubscriptionContainer &&
                        this.handleSelectors(subscriptionContainer, job);
                // Add Subscription Container to the 'subscriptionsToUpdate' queue
                if (updateSubscriptionContainer) {
                    subscriptionContainer.updatedSubscribers.add(job.observer);
                    subscriptionsToUpdate.add(subscriptionContainer);
                }
                job.subscriptionContainersToUpdate.delete(subscriptionContainer);
            });
        }
        return Array.from(subscriptionsToUpdate);
    }
    /**
     * Updates the specified Subscription Containers.
     *
     * Updating a Subscription Container triggers a re-render
     * on the Component it represents, based on the type of the Subscription Containers.
     *
     * @internal
     * @param subscriptionsToUpdate - Subscription Containers to be updated.
     */
    updateSubscriptionContainer(subscriptionsToUpdate) {
        // Using for loop for performance optimization
        // https://stackoverflow.com/questions/43821759/why-array-foreach-is-slower-than-for-loop-in-javascript
        for (let i = 0; i < subscriptionsToUpdate.length; i++) {
            const subscriptionContainer = subscriptionsToUpdate[i];
            // Call 'callback function' if Callback based Subscription
            if (subscriptionContainer instanceof CallbackSubscriptionContainer)
                subscriptionContainer.callback();
            // Call 'update method' in Integrations if Component based Subscription
            if (subscriptionContainer instanceof ComponentSubscriptionContainer)
                this.agileInstance().integrations.update(subscriptionContainer.component, this.getUpdatedObserverValues(subscriptionContainer));
            subscriptionContainer.updatedSubscribers.clear();
        }
        LogCodeManager.logIfTags(['runtime'], '16:01:02', [], subscriptionsToUpdate);
    }
    /**
     * Maps the values of the updated Observers (`updatedSubscribers`)
     * of the specified Subscription Container into a key map object.
     *
     * The key containing the Observer value is extracted from the Observer itself
     * or from the Subscription Container's `subscriberKeysWeakMap`.
     *
     * @internal
     * @param subscriptionContainer - Subscription Container from which the `updatedSubscribers` are to be mapped into a key map.
     */
    getUpdatedObserverValues(subscriptionContainer) {
        const props = {};
        for (const observer of subscriptionContainer.updatedSubscribers) {
            const key = subscriptionContainer.subscriberKeysWeakMap.get(observer) ??
                observer.key;
            if (key != null)
                props[key] = observer.value;
        }
        return props;
    }
    /**
     * Returns a boolean indicating whether the specified Subscription Container can be updated or not,
     * based on its selector functions (`selectorsWeakMap`).
     *
     * This is done by checking the '.value' and the '.previousValue' property of the Observer represented by the Job.
     * If a selected property differs, the Subscription Container (UI-Component) is allowed to update (re-render)
     * and `true` is returned.
     *
     * If the Subscription Container has no selector function at all, `true` is returned.
     *
     * @internal
     * @param subscriptionContainer - Subscription Container to be checked if it can be updated.
     * @param job - Job containing the Observer that is subscribed to the Subscription Container.
     */
    handleSelectors(subscriptionContainer, job) {
        const selectorMethods = subscriptionContainer.selectorsWeakMap.get(job.observer)?.methods;
        // If no selector functions found, return true.
        // Because no specific part of the Observer was selected.
        // -> The Subscription Container should be updated
        //    no matter what has updated in the Observer.
        if (selectorMethods == null)
            return true;
        // Check if a selected part of the Observer value has changed
        const previousValue = job.observer.previousValue;
        const newValue = job.observer.value;
        for (const selectorMethod of selectorMethods) {
            if (utils.notEqual(selectorMethod(newValue), selectorMethod(previousValue))
            // || newValueDeepness !== previousValueDeepness // Not possible to check the object deepness
            )
                return true;
        }
        return false;
    }
}

class RuntimeJob {
    config;
    // Key/Name identifier of the Runtime Job
    _key;
    // Observer the Job represents
    observer;
    // Whether the Subscription Containers (UI-Components) of the Observer should be updated (re-rendered)
    rerender;
    // Subscription Containers (UI-Components) of the Observer that have to be updated (re-rendered)
    subscriptionContainersToUpdate = new Set();
    // How often not ready Subscription Containers of the Observer have been tried to update
    timesTriedToUpdateCount = 0;
    // Whether the Job has been performed by the runtime
    performed = false;
    /**
     * A Runtime Job is sent to the Runtime on behalf of the Observer it represents.
     *
     * In the Runtime, the Observer is performed via its `perform()` method
     * and the Subscription Containers (UI-Components)
     * to which it is subscribed are updated (re-rendered) accordingly.
     *
     * @internal
     * @param observer - Observer to be represented by the Runtime Job.
     * @param config - Configuration object
     */
    constructor(observer, config = {}) {
        config = utils.defineConfig(config, {
            background: false,
            sideEffects: {
                enabled: true,
                exclude: [],
            },
            force: false,
            maxTriesToUpdate: 3,
            any: {},
        });
        this.config = {
            background: config.background,
            force: config.force,
            sideEffects: config.sideEffects,
            maxTriesToUpdate: config.maxTriesToUpdate,
            any: config.any,
        };
        this.observer = observer;
        this.rerender =
            !config.background &&
                this.observer.agileInstance().integrations.hasIntegration();
        this._key = config.key;
        this.subscriptionContainersToUpdate = new Set(observer.subscribedTo);
    }
    /**
     * Updates the key/name identifier of the Runtime Job.
     *
     * @public
     * @param value - New key/name identifier.
     */
    set key(value) {
        this._key = value;
    }
    /**
     * Returns the key/name identifier of the Runtime Job.
     *
     * @public
     */
    get key() {
        return this._key;
    }
}

class Observer {
    // Agile Instance the Observer belongs to
    agileInstance;
    // Key/Name identifier of the Observer
    _key;
    // Observers that depend on this Observer
    dependents = new Set();
    // Subscription Containers (UI-Components) the Observer is subscribed to
    subscribedTo = new Set();
    // Current value of the Observer
    value;
    // Previous value of the Observer
    previousValue;
    /**
     * An Observer manages the subscriptions to Subscription Containers (UI-Components)
     * and dependencies to other Observers (Agile Classes)
     * for an Agile Class such as the `State Class`.
     *
     * Agile Classes often use an Observer as an interface to the Runtime.
     * In doing so, they ingest their own Observer into the Runtime
     * when the Agile Class has changed in such a way
     * that these changes need to be applied to UI-Components
     * or dependent other Observers.
     *
     * After the Observer has been ingested into the Runtime
     * wrapped into a Runtime-Job, it is first added to the Jobs queue
     * to prevent race conditions.
     * When it is executed, the Observer's `perform()` method is called,
     * where the accordingly changes are applied to the Agile Class.
     *
     * Now that the Job was performed, it is added to the rerender queue,
     * where the subscribed Subscription Container (UI-Components)
     * of the Observer are updated (re-rendered).
     *
     * Note that the Observer itself is no standalone class
     * and should be adapted to the Agile Class needs it belongs to.
     *
     * @internal
     * @param agileInstance - Instance of Agile the Observer belongs to.
     * @param config - Configuration object
     */
    constructor(agileInstance, config = {}) {
        config = utils.defineConfig(config, {
            dependents: [],
            subs: [],
        });
        this.agileInstance = () => agileInstance;
        this._key = config.key;
        this.value = config.value;
        this.previousValue = config.value;
        config.dependents?.forEach((observer) => this.addDependent(observer));
        config.subs?.forEach((subscriptionContainer) => subscriptionContainer.addSubscription(this));
    }
    /**
     * Updates the key/name identifier of the Observer.
     *
     * @public
     * @param value - New key/name identifier.
     */
    set key(value) {
        this._key = value;
    }
    /**
     * Returns the key/name identifier of the Observer.
     *
     * @public
     */
    get key() {
        return this._key;
    }
    /**
     * Passes the Observer into the runtime wrapped into a Runtime-Job
     * where it is executed accordingly.
     *
     * During the execution the runtime performs the Observer's `perform()` method,
     * updates its dependents and re-renders the UI-Components it is subscribed to.
     *
     * @public
     * @param config - Configuration object
     */
    ingest(config = {}) {
        config = utils.defineConfig(config, {
            perform: true,
            background: false,
            sideEffects: {
                enabled: true,
                exclude: [],
            },
            force: false,
        });
        // Create Runtime-Job
        const job = new RuntimeJob(this, {
            force: config.force,
            sideEffects: config.sideEffects,
            background: config.background,
            key: config.key ??
                `${this._key != null ? this._key + '_' : ''}${utils.generateId()}`,
        });
        // Pass created Job into the Runtime
        this.agileInstance().runtime.ingest(job, {
            perform: config.perform,
        });
    }
    /**
     * Method executed by the Runtime to perform the Runtime-Job,
     * previously ingested via the `ingest()` method.
     *
     * Note that this method should be overwritten
     * to correctly apply the changes to the Agile Class
     * the Observer belongs to.
     *
     * @public
     * @param job - Runtime-Job to be performed.
     */
    perform(job) {
        LogCodeManager.log('17:03:00');
    }
    /**
     * Makes the specified Observer depend on the Observer.
     *
     * A dependent Observer is always ingested into the Runtime,
     * when the Observer it depends on has been ingested too.
     *
     * @public
     * @param observer - Observer to depend on the Observer.
     */
    addDependent(observer) {
        if (!this.dependents.has(observer))
            this.dependents.add(observer);
    }
    /**
     * Makes the specified Observer no longer depend on the Observer.
     *
     * @public
     * @param observer - Observer to no longer depend on the Observer.
     */
    removeDependent(observer) {
        if (this.dependents.has(observer))
            this.dependents.delete(observer);
    }
}

function extractObservers(instances) {
    const observers = [];
    const tempInstancesArray = utils.normalizeArray(instances, {
        createUndefinedArray: true,
    });
    // Extract Observers from specified Instances
    for (const instance of tempInstancesArray) {
        // If the Instance equals to 'null' or 'undefined'
        if (instance == null) {
            observers.push({});
            continue;
        }
        // If the Instance equals to a Collection
        if (instance.isCollection) {
            observers.push(instance.getGroupWithReference(instance.config.defaultGroupKey)
                .observers);
            continue;
        }
        // If the Instance contains a property that is an Observer
        if (instance['observer'] && instance['observer'] instanceof Observer) {
            observers.push({ value: instance['observer'] });
            continue;
        }
        // If the Instance contains a property that contains multiple Observers
        if (instance['observers']) {
            const extractedObservers = {};
            for (const key in instance['observers']) {
                if (instance['observers'][key] instanceof Observer) {
                    extractedObservers[key] = instance['observers'][key];
                }
            }
            observers.push(extractedObservers);
            continue;
        }
        // If the Instance equals to an Observer
        if (instance instanceof Observer) {
            observers.push({ value: instance });
            continue;
        }
        // Push empty object if no valid Observer was found
        observers.push({});
    }
    return Array.isArray(instances) ? observers : observers[0];
}
function extractRelevantObservers(instances, observerType) {
    const depsWithIndicator = {};
    const depsWithNoIndicator = [];
    // Extract Observers from deps
    for (const depKey in instances) {
        const extractedObservers = extractObservers(instances[depKey]);
        let observer = undefined;
        // Extract Observer at specified type
        if (observerType != null && extractedObservers[observerType] != null)
            observer = extractedObservers[observerType];
        // Extract most relevant Observer
        if (observerType == null)
            observer = extractedObservers['output'] ?? extractedObservers['value'];
        if (Array.isArray(instances))
            depsWithNoIndicator.push(observer);
        else
            depsWithIndicator[depKey] = observer;
    }
    return Array.isArray(instances) ? depsWithNoIndicator : depsWithIndicator;
}
/**
 * Retrieves the module with the specified key/name identifier
 * and returns `null` if the module couldn't be found.
 *
 * @param moduleName - Key/Name identifier of the module to be retrieved.
 * @param error - Whether to print an error, when the module couldn't be retrieved.
 */
function optionalRequire(moduleName, error = true) {
    let requiredPackage = null;
    try {
        requiredPackage = require(moduleName);
    }
    catch (e) {
        if (error) {
            LogCodeManager.log('20:03:02', [moduleName]);
            console.log(e);
        }
    }
    return requiredPackage;
}
/**
 * Binds the specified Instance globally at the provided key identifier.
 *
 * Learn more about global bound instances:
 * https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/globalThis
 *
 * @public
 * @param key - Key/Name identifier of the specified Instance.
 * @param instance - Instance to be bound globally.
 * @param overwrite - When already an Instance exists globally at the specified key,
 * whether to overwrite it with the new Instance.
 */
function globalBind(key, instance, overwrite = false) {
    try {
        if (overwrite) {
            globalThis[key] = instance;
            return true;
        }
        if (globalThis[key] == null) {
            globalThis[key] = instance;
            return true;
        }
    }
    catch (e) {
        LogCodeManager.log('20:03:01', [key]);
    }
    return false;
}
/**
 * Returns a boolean indicating whether AgileTs is currently running on a server.
 *
 * @public
 */
const runsOnServer = () => {
    return !(typeof window !== 'undefined' &&
        typeof window.document !== 'undefined' &&
        typeof window.document.createElement !== 'undefined');
};

class Integration {
    // Key/Name identifier of the Integration
    _key;
    // Instance of the Framework the Integration represents
    frameworkInstance;
    // Whether the Integration is ready and the binding to AgileTs was successful
    ready = false;
    // Whether the Integration was integrated into AgileTs
    integrated = false;
    // Methods to interact with the Framework represented by the Integration
    methods;
    /**
     * An Integration is an interface to a UI-Framework,
     * and allows the easy interaction with that Framework.
     *
     * Due to the Integration, AgileTs can be integrated into almost any UI-Framework
     * without a huge overhead.
     *
     * @public
     * @param config - Configuration object
     */
    constructor(config) {
        this._key = config.key;
        this.frameworkInstance = config.frameworkInstance;
        this.methods = {
            bind: config.bind,
            updateMethod: config.updateMethod,
        };
    }
    /**
     * Updates the key/name identifier of the Integration.
     *
     * @public
     * @param value - New key/name identifier.
     */
    set key(value) {
        this._key = value;
    }
    /**
     * Returns the key/name identifier of the Integration.
     *
     * @public
     */
    get key() {
        return this._key;
    }
}

const onRegisterInitialIntegrationCallbacks = [];
class Integrations {
    // Agile Instance the Integrations belongs to
    agileInstance;
    // Registered Integrations
    integrations = new Set();
    // External added Integrations
    // that are to integrate into not yet existing Agile Instances
    static initialIntegrations = [];
    /**
     * Registers the specified Integration in each existing or not-yet created Agile Instance.
     *
     * @public
     * @param integration - Integration to be registered in each Agile Instance.
     */
    static addInitialIntegration(integration) {
        if (integration instanceof Integration) {
            // Executed external registered Integration callbacks
            onRegisterInitialIntegrationCallbacks.forEach((callback) => callback(integration));
            Integrations.initialIntegrations.push(integration);
        }
    }
    /**
     * Fires on each external added Integration.
     *
     * @public
     * @param callback - Callback to be fired when an Integration was externally added.
     */
    static onRegisterInitialIntegration(callback) {
        onRegisterInitialIntegrationCallbacks.push(callback);
        Integrations.initialIntegrations.forEach((integration) => {
            callback(integration);
        });
    }
    /**
     * The Integrations Class manages all Integrations for an Agile Instance
     * and provides an interface to easily update
     * and invoke functions in all registered Integrations.
     *
     * @internal
     * @param agileInstance - Instance of Agile the Integrations belongs to.
     * @param config - Configuration object
     */
    constructor(agileInstance, config = {}) {
        config = utils.defineConfig(config, {
            autoIntegrate: true,
        });
        this.agileInstance = () => agileInstance;
        if (config.autoIntegrate) {
            // Setup listener to be notified when an external registered Integration was added
            Integrations.onRegisterInitialIntegration((integration) => {
                this.integrate(integration);
            });
        }
    }
    /**
     * Integrates the specified Integration into AgileTs
     * and sets it to ready when the binding was successful.
     *
     * @public
     * @param integration - Integration to be integrated into AgileTs.
     */
    async integrate(integration) {
        // Check if Integration is valid
        if (integration._key == null) {
            LogCodeManager.log('18:03:00', [integration._key, this.agileInstance().key], integration);
            return false;
        }
        // Bind to integrate Integration to AgileTs
        if (integration.methods.bind)
            integration.ready = await integration.methods.bind(this.agileInstance());
        else
            integration.ready = true;
        // Integrate Integration
        this.integrations.add(integration);
        integration.integrated = true;
        LogCodeManager.log('18:00:00', [integration._key, this.agileInstance().key], integration);
        return true;
    }
    /**
     * Updates the specified UI-Component Instance
     * with the updated data object in all registered Integrations that are ready.
     *
     * In doing so, it calls the `updateMethod()` method
     * in all registered Integrations with the specified parameters.
     *
     * @public
     * @param componentInstance - Component Instance to be updated.
     * @param updatedData - Data object with updated data.
     */
    update(componentInstance, updatedData) {
        this.integrations.forEach((integration) => {
            if (!integration.ready) {
                LogCodeManager.log('18:02:00', [integration._key]);
                return;
            }
            if (integration.methods.updateMethod)
                integration.methods.updateMethod(componentInstance, updatedData);
        });
    }
    /**
     * Returns a boolean indicating whether any Integration
     * has been registered with the Agile Instance or not.
     *
     * @public
     */
    hasIntegration() {
        return this.integrations.size > 0;
    }
}

class Agile$1 {
    config;
    // Key/Name identifier of Agile Instance
    key;
    // Queues and executes incoming Observer-based Jobs
    runtime;
    // Manages and simplifies the subscription to UI-Components
    subController;
    // Integrations (UI-Frameworks) that are integrated into the Agile Instance
    integrations;
    // Identifier used to bind an Agile Instance globally
    static globalKey = '__agile__';
    /**
     * The Agile Class is the main Instance of AgileTs
     * and should be unique to your application.
     *
     * Simply put, the Agile Instance is the brain of AgileTs
     * and manages all [Agile Sub Instance](https://agile-ts.org/docs/introduction/#agile-sub-instance)
     * such as States.
     *
     * It should be noted that it doesn't store the States;
     * It only manages them. Each State has an Instance of the Agile Class,
     * for example, to ingest its changes into the Runtime.
     * In summary, the main tasks of the Agile Class are to:
     * - queue [Agile Sub Instance](https://agile-ts.org/docs/introduction/#agile-sub-instance)
     *   changes in the Runtime to prevent race conditions
     * - update/rerender subscribed UI-Components through the provided Integrations
     *   such as the [React Integration](https://agile-ts.org/docs/react)
     * - provide configuration object
     *
     * Each Agile Sub Instance requires an Agile Instance to be instantiated and function properly.
     *
     * [Learn more..](https://agile-ts.org/docs/core/agile-instance/)
     *
     * @public
     * @param config - Configuration object
     */
    constructor(config = {}) {
        config = utils.defineConfig(config, {
            localStorage: false,
            waitForMount: true,
            bindGlobal: false,
            autoIntegrate: true,
            bucket: true,
        });
        this.config = {
            waitForMount: config.waitForMount,
            bucket: config.bucket,
        };
        this.key = config.key;
        this.integrations = new Integrations(this, {
            autoIntegrate: config.autoIntegrate,
        });
        this.runtime = new Runtime(this);
        this.subController = new SubController(this);
        LogCodeManager.log('10:00:00', [], this);
        // Create a global instance of the Agile Instance.
        // Why? 'getAgileInstance()' returns the global Agile Instance
        // if it couldn't find any Agile Instance in the specified Instance.
        if (config.bindGlobal)
            if (!globalBind(Agile$1.globalKey, this)) {
                LogCodeManager.log('10:02:00');
            }
    }
    /**
     * Registers the specified Integration with AgileTs.
     *
     * After a successful registration,
     * [Agile Sub Instances](https://agile-ts.org/docs/introduction/#agile-sub-instance) such as States
     * can be bound to the Integration's UI-Components for reactivity.
     *
     * [Learn more..](https://agile-ts.org/docs/core/agile-instance/methods#integrate)
     *
     * @public
     * @param integration - Integration to be integrated/registered.
     */
    integrate(integration) {
        this.integrations.integrate(integration);
        return this;
    }
    /**
     * Returns a boolean indicating whether any Integration
     * has been registered with AgileTs or not.
     *
     * [Learn more..](https://agile-ts.org/docs/core/agile-instance/methods#hasintegration)
     *
     * @public
     */
    hasIntegration() {
        return this.integrations.hasIntegration();
    }
}

class Storage {
    config;
    // Key/Name identifier of the Storage
    key;
    // Whether the Storage is ready and is able to store values
    ready = false;
    // Methods to interact with the external Storage (get, set, remove)
    methods;
    /**
     * A Storage is an interface to an external Storage,
     * and allows the easy interaction with that Storage.
     *
     * Due to the Storage, AgileTs can easily persist its Instances in almost any Storage
     * without a huge overhead.
     *
     * @public
     * @param config - Configuration object
     */
    constructor(config) {
        config = utils.defineConfig(config, {
            prefix: 'agile',
            async: false,
        });
        this.key = config.key;
        this.methods = config.methods;
        this.config = {
            prefix: config.prefix,
            async: config.async,
        };
        this.ready = this.validate();
        if (!this.ready)
            return;
        // Check if Storage is async
        if (utils.isAsyncFunction(this.methods.get) ||
            utils.isAsyncFunction(this.methods.set) ||
            utils.isAsyncFunction(this.methods.remove))
            this.config.async = true;
    }
    /**
     * Returns a boolean indicating whether the Storage is valid
     * and can be used to persist Instances in it or not.
     *
     * @public
     */
    validate() {
        if (!utils.isFunction(this.methods?.get)) {
            LogCodeManager.log('13:03:00', ['get']);
            return false;
        }
        if (!utils.isFunction(this.methods?.set)) {
            LogCodeManager.log('13:03:00', ['set']);
            return false;
        }
        if (!utils.isFunction(this.methods?.remove)) {
            LogCodeManager.log('13:03:00', ['remove']);
            return false;
        }
        return true;
    }
    /**
     * Synchronously retrieves the stored value
     * at the specified Storage Item key from the external Storage.
     *
     * When the retrieved value is a JSON-String it is parsed automatically.
     *
     * @public
     * @param key - Key/Name identifier of the value to be retrieved.
     */
    normalGet(key) {
        if (!this.ready || !this.methods.get)
            return undefined;
        if (utils.isAsyncFunction(this.methods.get))
            LogCodeManager.log('13:02:00');
        // Retrieve value
        const res = this.methods.get(this.getStorageKey(key));
        const _res = utils.isJsonString(res) ? JSON.parse(res) : res;
        LogCodeManager.logIfTags(['storage'], '13:01:00', [this.key, this.getStorageKey(key)], _res);
        return _res;
    }
    /**
     * Asynchronously retrieves the stored value
     * at the specified Storage Item key from the external Storage.
     *
     * When the retrieved value is a JSON-String it is parsed automatically.
     *
     * @public
     * @param key - Key/Name identifier of the value to be retrieved.
     */
    get(key) {
        if (!this.ready || !this.methods.get)
            return Promise.resolve(undefined);
        // Retrieve value from not promise based Storage
        if (!utils.isAsyncFunction(this.methods.get))
            return Promise.resolve(this.normalGet(key));
        // Retrieve value from promise based Storage
        return new Promise((resolve, reject) => {
            this.methods
                ?.get(this.getStorageKey(key))
                .then((res) => {
                const _res = utils.isJsonString(res) ? JSON.parse(res) : res;
                LogCodeManager.logIfTags(['storage'], '13:01:00', [this.key, this.getStorageKey(key)], _res);
                resolve(_res);
            })
                .catch(reject);
        });
    }
    /**
     * Stores or updates the value at the specified Storage Item key
     * in the external Storage.
     *
     * @public
     * @param key - Key/Name identifier of the value to be stored or updated.
     * @param value - Value to be stored.
     */
    set(key, value) {
        if (!this.ready || !this.methods.set)
            return;
        LogCodeManager.logIfTags(['storage'], '13:01:01', [
            this.key,
            this.getStorageKey(key),
        ]);
        this.methods.set(this.getStorageKey(key), JSON.stringify(value));
    }
    /**
     * Removes the value at the specified Storage Item key
     * from the external Storage.
     *
     * @public
     * @param key - Key/Name identifier of the value to be removed.
     */
    remove(key) {
        if (!this.ready || !this.methods.remove)
            return;
        LogCodeManager.logIfTags(['storage'], '13:01:02', [
            this.key,
            this.getStorageKey(key),
        ]);
        this.methods.remove(this.getStorageKey(key));
    }
    /**
     * Generates and returns a valid Storage key based on the specified key.
     *
     * @internal
     * @param key - Key to be converted into a valid Storage key.
     */
    getStorageKey(key) {
        return this.config.prefix
            ? `_${this.config.prefix}_${key}`
            : key.toString();
    }
}

class Storages {
    // Agile Instance the Storages belongs to
    agileInstance;
    config;
    // Registered Storages
    storages = {};
    // Persistent from Instances (for example States) that were persisted
    persistentInstances = {};
    /**
     * The Storages Class manages all external Storages for an Agile Instance
     * and provides an interface to easily store,
     * load and remove values from multiple Storages at once.
     *
     * @internal
     * @param agileInstance - Instance of Agile the Storages belongs to.
     * @param config - Configuration object
     */
    constructor(agileInstance, config = {}) {
        this.agileInstance = () => agileInstance;
        config = utils.defineConfig(config, {
            localStorage: false,
            defaultStorageKey: null,
        });
        this.config = { defaultStorageKey: config.defaultStorageKey };
        if (config.localStorage)
            this.instantiateLocalStorage();
    }
    /**
     * Instantiates and registers the
     * [Local Storage](https://developer.mozilla.org/de/docs/Web/API/Window/localStorage).
     *
     * Note that the Local Storage is only available in a web environment.
     *
     * @internal
     */
    instantiateLocalStorage() {
        if (!Storages.localStorageAvailable()) {
            LogCodeManager.log('11:02:00');
            return false;
        }
        const _localStorage = new Storage({
            key: 'localStorage',
            async: false,
            methods: {
                get: localStorage.getItem.bind(localStorage),
                set: localStorage.setItem.bind(localStorage),
                remove: localStorage.removeItem.bind(localStorage),
            },
        });
        return this.register(_localStorage, { default: true });
    }
    /**
     * Registers the specified Storage with AgileTs
     * and updates the Persistent Instances that have already attempted
     * to use the previously unregistered Storage.
     *
     * @public
     * @param storage - Storage to be registered with AgileTs.
     * @param config - Configuration object
     */
    register(storage, config = {}) {
        const hasRegisteredAnyStorage = utils.notEqual(this.storages, {});
        // Check if Storage already exists
        if (Object.prototype.hasOwnProperty.call(this.storages, storage.key)) {
            LogCodeManager.log('11:03:00', [storage.key]);
            return false;
        }
        // Assign Storage as default Storage if it is the first one added
        if (!hasRegisteredAnyStorage && config.default === false)
            LogCodeManager.log('11:02:01');
        if (!hasRegisteredAnyStorage)
            config.default = true;
        // Register Storage
        this.storages[storage.key] = storage;
        if (config.default)
            this.config.defaultStorageKey = storage.key;
        for (const persistentKey of Object.keys(this.persistentInstances)) {
            const persistent = this.persistentInstances[persistentKey];
            if (persistent == null)
                continue;
            // Revalidate Persistent, which contains key/name identifier of the newly registered Storage
            if (persistent.storageKeys.includes(storage.key)) {
                const isValid = persistent.validatePersistent();
                if (isValid)
                    persistent.initialLoading();
                continue;
            }
            // If Persistent has no default Storage key,
            // reassign Storage keys since the now registered Storage
            // might be tagged as default Storage of AgileTs
            if (persistent.config.defaultStorageKey == null) {
                persistent.assignStorageKeys();
                const isValid = persistent.validatePersistent();
                if (isValid)
                    persistent.initialLoading();
            }
        }
        LogCodeManager.log('13:00:00', [storage.key], storage);
        return true;
    }
    /**
     * Retrieves a single Storage with the specified key/name identifier
     * from the Storages Class.
     *
     * If the to retrieve Storage doesn't exist, `undefined` is returned.
     *
     * @public
     * @param storageKey - Key/Name identifier of the Storage.
     */
    getStorage(storageKey) {
        if (!storageKey)
            return undefined;
        const storage = this.storages[storageKey];
        if (!storage) {
            LogCodeManager.log('11:03:01', [storageKey]);
            return undefined;
        }
        if (!storage.ready) {
            LogCodeManager.log('11:03:02', [storageKey]);
            return undefined;
        }
        return storage;
    }
    /**
     * Retrieves the stored value at the specified Storage Item key
     * from the defined external Storage (`storageKey`).
     *
     * When no Storage has been specified,
     * the value is retrieved from the default Storage.
     *
     * @public
     * @param storageItemKey - Key/Name identifier of the value to be retrieved.
     * @param storageKey - Key/Name identifier of the external Storage
     * from which the value is to be retrieved.
     */
    get(storageItemKey, storageKey) {
        if (!this.hasStorage()) {
            LogCodeManager.log('11:03:03');
            return Promise.resolve(undefined);
        }
        // Call get method on specified Storage
        if (storageKey) {
            const storage = this.getStorage(storageKey);
            if (storage)
                return storage.get(storageItemKey);
        }
        // Call get method on default Storage
        const defaultStorage = this.getStorage(this.config.defaultStorageKey);
        return (defaultStorage?.get(storageItemKey) || Promise.resolve(undefined));
    }
    /**
     * Stores or updates the value at the specified Storage Item key
     * in the defined external Storages (`storageKeys`).
     *
     * When no Storage has been specified,
     * the value is stored/updated in the default Storage.
     *
     * @public
     * @param storageItemKey - Key/Name identifier of the value to be stored.
     * @param value - Value to be stored in an external Storage.
     * @param storageKeys - Key/Name identifier of the external Storage
     * where the value is to be stored.
     */
    set(storageItemKey, value, storageKeys) {
        if (!this.hasStorage()) {
            LogCodeManager.log('11:03:04');
            return;
        }
        // Call set method on specified Storages
        if (storageKeys != null) {
            for (const storageKey of storageKeys)
                this.getStorage(storageKey)?.set(storageItemKey, value);
            return;
        }
        // Call set method on default Storage
        const defaultStorage = this.getStorage(this.config.defaultStorageKey);
        defaultStorage?.set(storageItemKey, value);
    }
    /**
     * Removes the value at the specified Storage Item key
     * from the defined external Storages (`storageKeys`).
     *
     * When no Storage has been specified,
     * the value is removed from the default Storage.
     *
     * @public
     * @param storageItemKey - Key/Name identifier of the value to be removed.
     * @param storageKeys - Key/Name identifier of the external Storage
     * from which the value is to be removed.
     */
    remove(storageItemKey, storageKeys) {
        if (!this.hasStorage()) {
            LogCodeManager.log('11:03:05');
            return;
        }
        // Call remove method on specified Storages
        if (storageKeys) {
            for (const storageKey of storageKeys)
                this.getStorage(storageKey)?.remove(storageItemKey);
            return;
        }
        // Call remove method on default Storage
        const defaultStorage = this.getStorage(this.config.defaultStorageKey);
        defaultStorage?.remove(storageItemKey);
    }
    /**
     * Returns a boolean indicating whether any Storage
     * has been registered with the Agile Instance or not.
     *
     * @public
     */
    hasStorage() {
        return utils.notEqual(this.storages, {});
    }
    /**
     * Returns a boolean indication whether the
     * [Local Storage](https://developer.mozilla.org/de/docs/Web/API/Window/localStorage)
     * is available in the current environment.
     *
     * @public
     */
    static localStorageAvailable() {
        try {
            localStorage.setItem('_myDummyKey_', 'myDummyValue');
            localStorage.removeItem('_myDummyKey_');
            return true;
        }
        catch (e) {
            return false;
        }
    }
}

/**
 * Shared Agile Instance that is used when no Agile Instance was specified.
 */
exports.sharedAgileInstance = new Agile$1({
    key: 'shared',
});
/**
 * Assigns the specified Agile Instance as the shared Agile Instance.
 *
 * @param agileInstance - Agile Instance to become the new shared Agile Instance.
 */
// https://stackoverflow.com/questions/32558514/javascript-es6-export-const-vs-export-let
function assignSharedAgileInstance(agileInstance) {
    exports.sharedAgileInstance = agileInstance;
}
// Note: Not located in 'utils.ts' due circular dependency issues
/**
 * Extracts an Instance of Agile from the specified Instance.
 * When no valid Agile Instance was found,
 * it returns the global bound Agile Instance or `undefined`.
 *
 * @internal
 * @param instance - Instance to extract the Agile Instance from.
 */
function getAgileInstance(instance) {
    try {
        // Try to get Agile Instance from specified Instance
        if (instance != null) {
            const _agileInstance = utils.isFunction(instance['agileInstance'])
                ? instance['agileInstance']()
                : instance['agileInstance'];
            if (_agileInstance)
                return _agileInstance;
        }
        // Try to get shared Agile Instance
        if (exports.sharedAgileInstance instanceof Agile$1) {
            return exports.sharedAgileInstance;
        }
        // Return global bound Agile Instance
        return globalThis[Agile$1.globalKey];
    }
    catch (e) {
        LogCodeManager.log('20:03:00', [], instance);
    }
    return undefined;
}

/**
 * Handles the permanent persistence of Agile Classes.
 */
exports.sharedStorageManager = null;
/**
 * Assigns the specified Storage Manager
 * as default (shared) Storage Manager for all Agile Instances.
 *
 *  @param instance - Storage Manager to be assigned as the shared Storage Manager.
 */
// https://stackoverflow.com/questions/32558514/javascript-es6-export-const-vs-export-let
const assignSharedStorageManager = (instance) => {
    exports.sharedStorageManager = instance;
};
/**
 * Returns a newly created Storage Manager.
 *
 * A Storage Manager manages all external Storages for AgileTs
 * and provides an interface to easily store,
 * load and remove values from multiple external Storages at once.
 *
 * @param config - Configuration object
 */
function createStorageManager(config = {}) {
    config = utils.defineConfig(config, {
        agileInstance: exports.sharedAgileInstance,
    });
    return new Storages(config.agileInstance, utils.removeProperties(config, ['agileInstance']));
}
/**
 * Returns the shared Storage Manager
 * or creates a new one when no shared Storage Manager exists.
 */
function getSharedStorageManager() {
    if (exports.sharedStorageManager == null) {
        const newStorageManager = createStorageManager({
            localStorage: !runsOnServer(),
        });
        assignSharedStorageManager(newStorageManager);
        return newStorageManager;
    }
    return exports.sharedStorageManager;
}

class Persistent {
    // Agile Instance the Persistent belongs to
    agileInstance;
    static placeHolderKey = '__THIS_IS_A_PLACEHOLDER__';
    config;
    // Key/Name identifier of the Persistent
    _key;
    // Whether the Persistent is ready
    // and is able to persist values in an external Storage
    ready = false;
    // Whether the Persistent value is stored in a corresponding external Storage/s
    isPersisted = false;
    // Callback that is called when the persisted value was loaded into the Persistent for the first time
    onLoad;
    // Key/Name identifier of the Storages the Persistent value is stored in
    storageKeys = [];
    /**
     * A Persistent manages the permanent persistence
     * of an Agile Class such as the `State Class` in external Storages.
     *
     * Note that the Persistent itself is no standalone class
     * and should be adapted to the Agile Class needs it belongs to.
     *
     * @internal
     * @param agileInstance - Instance of Agile the Persistent belongs to.
     * @param config - Configuration object
     */
    constructor(agileInstance, config = {}) {
        this.agileInstance = () => agileInstance;
        this._key = Persistent.placeHolderKey;
        config = utils.defineConfig(config, {
            loadValue: true,
            storageKeys: [],
            defaultStorageKey: null,
        });
        this.config = { defaultStorageKey: config.defaultStorageKey };
        // Instantiate Persistent
        if (config.loadValue) {
            this.instantiatePersistent({
                storageKeys: config.storageKeys,
                key: config.key,
                defaultStorageKey: config.defaultStorageKey,
            });
        }
    }
    /**
     * Updates the key/name identifier of the Persistent.
     *
     * @public
     * @param value - New key/name identifier.
     */
    set key(value) {
        this.setKey(value);
    }
    /**
     * Returns the key/name identifier of the Persistent.
     *
     * @public
     */
    get key() {
        return this._key;
    }
    /**
     * Updates the key/name identifier of the Persistent.
     *
     * @public
     * @param value - New key/name identifier.
     */
    async setKey(value) {
        const oldKey = this._key;
        const wasReady = this.ready;
        // Assign new key to Persistent
        if (value === this._key)
            return;
        this._key = value ?? Persistent.placeHolderKey;
        const isValid = this.validatePersistent();
        // Try to initial load value if Persistent hasn't been ready before
        if (!wasReady) {
            if (isValid)
                await this.initialLoading();
            return;
        }
        // Remove persisted value that is located at the old key
        await this.removePersistedValue(oldKey);
        // Persist value at the new key
        if (isValid)
            await this.persistValue(value);
    }
    /**
     * Instantiates the Persistent by assigning the specified Storage keys to it
     * and validating it to make sure everything was setup correctly.
     *
     * This was moved out of the `constructor()`
     * because some classes (that extend the Persistent) need to configure some
     * things before they can properly instantiate the parent Persistent.
     *
     * @internal
     * @param config - Configuration object
     */
    instantiatePersistent(config = {}) {
        this._key = this.formatKey(config.key) ?? Persistent.placeHolderKey;
        this.assignStorageKeys(config.storageKeys, config.defaultStorageKey);
        this.validatePersistent();
        // Register Persistent to Storage Manager
        const storageManager = getSharedStorageManager();
        if (this._key !== Persistent.placeHolderKey && storageManager != null) {
            storageManager.persistentInstances[this._key] = this;
        }
    }
    /**
     * Returns a boolean indicating whether the Persistent was setup correctly
     * and is able to persist a value permanently in an external Storage.
     *
     * Based on the tapped boolean value,
     * the Persistent's `ready` property is updated.
     *
     * @internal
     */
    validatePersistent() {
        let isValid = true;
        // Validate Persistent key/name identifier
        if (this._key === Persistent.placeHolderKey) {
            LogCodeManager.log('12:03:00');
            isValid = false;
        }
        // Validate Storage keys
        if (this.config.defaultStorageKey == null || this.storageKeys.length <= 0) {
            LogCodeManager.log('12:03:01');
            isValid = false;
        }
        // Check if the Storages exist at the specified Storage keys
        this.storageKeys.map((key) => {
            if (!getSharedStorageManager()?.storages[key]) {
                LogCodeManager.log('12:03:02', [this._key, key]);
                isValid = false;
            }
        });
        this.ready = isValid;
        return isValid;
    }
    /**
     * Assigns the specified Storage identifiers to the Persistent
     * and extracts the default Storage if necessary.
     *
     * When no Storage key was provided the default Storage
     * of the Agile Instance is applied to the Persistent.
     *
     * @internal
     * @param storageKeys - Key/Name identifier of the Storages to be assigned.
     * @param defaultStorageKey - Key/Name identifier of the default Storage.
     */
    assignStorageKeys(storageKeys = [], defaultStorageKey) {
        const _storageKeys = utils.copy(storageKeys);
        // Assign specified default Storage key to the 'storageKeys' array
        if (defaultStorageKey != null && !_storageKeys.includes(defaultStorageKey))
            _storageKeys.push(defaultStorageKey);
        // Assign the default Storage key of the Agile Instance to the 'storageKeys' array
        // and specify it as the Persistent's default Storage key
        // if no valid Storage key was provided
        if (_storageKeys.length <= 0) {
            const defaultStorageKey = getSharedStorageManager()?.config
                .defaultStorageKey;
            if (defaultStorageKey != null) {
                this.config.defaultStorageKey = defaultStorageKey;
                _storageKeys.push(getSharedStorageManager()?.config.defaultStorageKey);
            }
        }
        else {
            this.config.defaultStorageKey = defaultStorageKey ?? _storageKeys[0];
        }
        this.storageKeys = _storageKeys;
    }
    /**
     * Stores or loads the Persistent value
     * from the external Storages for the first time.
     *
     * @internal
     */
    async initialLoading() {
        const success = await this.loadPersistedValue();
        if (this.onLoad)
            this.onLoad(success);
        if (!success)
            await this.persistValue();
    }
    /**
     * Loads the Persistent value from the corresponding Storage.
     *
     * Note that this method should be overwritten
     * to correctly apply the changes to the Agile Class
     * the Persistent belongs to.
     *
     * @internal
     * @param storageItemKey - Storage key of the to load value.
     * | default = Persistent.key |
     * @return Whether the loading of the persisted value was successful.
     */
    async loadPersistedValue(storageItemKey) {
        LogCodeManager.log('00:03:00', ['loadPersistedValue', 'Persistent']);
        return false;
    }
    /**
     * Persists the Persistent value in the corresponding Storage.
     *
     * Note that this method should be overwritten
     * to correctly apply the changes to the Agile Class
     * the Persistent belongs to.
     *
     * @internal
     * @param storageItemKey - Storage key of the to persist value
     * | default = Persistent.key |
     * @return Whether the persisting of the value was successful.
     */
    async persistValue(storageItemKey) {
        LogCodeManager.log('00:03:00', ['persistValue', 'Persistent']);
        return false;
    }
    /**
     * Removes the Persistent value from the corresponding Storage.
     * -> Persistent value is no longer persisted
     *
     * Note that this method should be overwritten
     * to correctly apply the changes to the Agile Class
     * the Persistent belongs to.
     *
     * @internal
     * @param storageItemKey - Storage key of the to remove value.
     * | default = Persistent.key |
     * @return Whether the removal of the persisted value was successful.
     */
    async removePersistedValue(storageItemKey) {
        LogCodeManager.log('00:03:00', ['removePersistedValue', 'Persistent']);
        return false;
    }
    /**
     * Formats the specified key so that it can be used as a valid Storage key
     * and returns the formatted variant of it.
     *
     * Note that this method should be overwritten
     * to correctly apply the changes to the Agile Class
     * the Persistent belongs to.
     *
     * @internal
     * @param key - Storage key to be formatted.
     */
    formatKey(key) {
        return key;
    }
}

/**
 * Returns a newly created Storage.
 *
 * A Storage Class serves as an interface to external storages,
 * such as the [Async Storage](https://github.com/react-native-async-storage/async-storage) or
 * [Local Storage](https://www.w3schools.com/html/html5_webstorage.asp).
 *
 * It creates the foundation to easily [`persist()`](https://agile-ts.org/docs/core/state/methods#persist) [Agile Sub Instances](https://agile-ts.org/docs/introduction/#agile-sub-instance)
 * (like States or Collections) in nearly any external storage.
 *
 * [Learn more..](https://agile-ts.org/docs/core/agile-instance/methods#createstorage)
 *
 * @public
 * @param config - Configuration object
 */
function createStorage(config) {
    return new Storage(config);
}

class StateRuntimeJob extends RuntimeJob {
    config;
    /**
     * A State Runtime Job is sent to the Runtime on behalf of the State Observer it represents.
     *
     * In the Runtime, the State Observer is performed via its `perform()` method
     * and the Subscription Containers (UI-Components)
     * to which it is subscribed are updated (re-rendered) accordingly.
     *
     * @internal
     * @param observer - State Observer to be represented by the State Runtime Job.
     * @param config - Configuration object
     */
    constructor(observer, config = {}) {
        super(observer, config);
        config = utils.defineConfig(config, {
            background: false,
            sideEffects: {
                enabled: true,
                exclude: [],
            },
            force: false,
            storage: true,
            overwrite: false,
            maxTriesToUpdate: 3,
            any: {},
        });
        this.config = {
            background: config.background,
            force: config.force,
            sideEffects: config.sideEffects,
            storage: config.storage,
            overwrite: config.overwrite,
            maxTriesToUpdate: config.maxTriesToUpdate,
            any: config.any,
        };
    }
}

class StateObserver extends Observer {
    // State the Observer belongs to
    state;
    // Next value applied to the State
    nextStateValue;
    /**
     * A State Observer manages the subscriptions to Subscription Containers (UI-Components)
     * and dependencies to other Observers (Agile Classes)
     * for a State Class.
     *
     * @internal
     * @param state - Instance of State the Observer belongs to.
     * @param config - Configuration object
     */
    constructor(state, config = {}) {
        super(state.agileInstance(), utils.defineConfig(config, {
            value: state._value,
        }));
        this.state = () => state;
        this.nextStateValue = utils.copy(state._value);
    }
    /**
     * Passes the State Observer into the runtime wrapped into a Runtime-Job
     * where it is executed accordingly.
     *
     * During the execution the runtime applies the `nextStateValue`
     * or the `computedValue` (Computed Class) to the State,
     * updates its dependents and re-renders the UI-Components it is subscribed to.
     *
     * @internal
     * @param config - Configuration object
     */
    ingest(config = {}) {
        const state = this.state();
        if (state.isComputed) {
            state.compute().then((result) => {
                this.ingestValue(result, config);
            });
        }
        else {
            this.ingestValue(state.nextStateValue, config);
        }
    }
    /**
     * Passes the State Observer into the runtime wrapped into a Runtime-Job
     * where it is executed accordingly.
     *
     * During the execution the runtime applies the specified `newStateValue` to the State,
     * updates its dependents and re-renders the UI-Components it is subscribed to.
     *
     * @internal
     * @param newStateValue - New value to be applied to the State.
     * @param config - Configuration object.
     */
    ingestValue(newStateValue, config = {}) {
        const state = this.state();
        config = utils.defineConfig(config, {
            perform: true,
            force: false,
        });
        // Force overwriting the State value if it is a placeholder.
        // After assigning a value to the State, the State is supposed to be no placeholder anymore.
        if (state.isPlaceholder) {
            config.force = true;
            config.overwrite = true;
        }
        // Assign next State value to Observer and compute it if necessary (enhanced State)
        this.nextStateValue = state.computeValueMethod
            ? utils.copy(state.computeValueMethod(newStateValue))
            : utils.copy(newStateValue);
        // Check if current State value and to assign State value are equal
        if (utils.equal(state._value, this.nextStateValue) && !config.force)
            return;
        // Create Runtime-Job
        const job = new StateRuntimeJob(this, {
            ...utils.removeProperties(config, ['perform']),
            ...{
                key: config.key ??
                    `${this._key != null ? this._key + '_' : ''}${utils.generateId()}_value`,
            },
        });
        // Pass created Job into the Runtime
        this.agileInstance().runtime.ingest(job, {
            perform: config.perform,
        });
    }
    /**
     * Method executed by the Runtime to perform the Runtime-Job,
     * previously ingested via the `ingest()` or `ingestValue()` method.
     *
     * Thereby the previously defined `nextStateValue` is assigned to the State.
     * Also side effects (like calling watcher callbacks) of a State change are executed.
     *
     * @internal
     * @param job - Runtime-Job to be performed.
     */
    perform(job) {
        const observer = job.observer;
        const state = observer.state();
        // Assign new State values
        state.previousStateValue = utils.copy(state._value);
        state._value = utils.copy(observer.nextStateValue);
        state.nextStateValue = utils.copy(observer.nextStateValue);
        // TODO think about freezing the State value..
        // https://www.geeksforgeeks.org/object-freeze-javascript/#:~:text=Object.freeze()%20Method&text=freeze()%20which%20is%20used,the%20prototype%20of%20the%20object.
        // if (typeof state._value === 'object') Object.freeze(state._value);
        // Overwrite entire State with the newly assigned value
        if (job.config.overwrite) {
            state.initialStateValue = utils.copy(state._value);
            state.previousStateValue = utils.copy(state._value);
            state.isPlaceholder = false;
        }
        state.isSet = utils.notEqual(state._value, state.initialStateValue);
        this.sideEffects(job);
        // Assign new public value to the Observer (value used by the Integrations)
        job.observer.previousValue = utils.copy(observer.value);
        job.observer.value = utils.copy(state._value);
    }
    /**
     * Performs the side effects of applying the next State value to the State.
     *
     * Side effects are, for example, calling the watcher callbacks
     * or executing the side effects defined in the State Class
     * like 'rebuildGroup' or 'rebuildStateStorageValue'.
     *
     * @internal
     * @param job - Job that is currently performed.
     */
    sideEffects(job) {
        const state = job.observer.state();
        // Call side effect functions
        if (job.config?.sideEffects?.enabled) {
            const sideEffectArray = utils.createArrayFromObject(state.sideEffects);
            sideEffectArray.sort(function (a, b) {
                return b.instance.weight - a.instance.weight;
            });
            for (const sideEffect of sideEffectArray) {
                if (utils.isFunction(sideEffect.instance.callback)) {
                    if (!job.config.sideEffects.exclude?.includes(sideEffect.key))
                        sideEffect.instance.callback(job.observer.state(), job.config);
                }
            }
        }
    }
}

class ComputedTracker {
    static isTracking = false;
    static trackedObservers = new Set();
    /**
     * Helper Class for automatic tracking used Observers (dependencies) in a compute function.
     *
     * @internal
     */
    constructor() {
        // empty
    }
    /**
     * Activates Computed Tracker to globally track used Observers.
     *
     * @internal
     */
    static track() {
        this.isTracking = true;
    }
    /**
     * Tracks the passed Observer and caches it
     * when the Computed Tracker is actively tracking.
     *
     * @internal
     * @param observer - Observer
     */
    static tracked(observer) {
        if (this.isTracking)
            this.trackedObservers.add(observer);
    }
    /**
     * Returns the latest tracked Observers
     * and stops the Computed Tracker from tracking any more Observers.
     *
     * @internal
     */
    static getTrackedObservers() {
        const trackedObservers = Array.from(this.trackedObservers);
        // Reset Computed Tracker
        this.isTracking = false;
        this.trackedObservers = new Set();
        return trackedObservers;
    }
}

class State {
    // Agile Instance the State belongs to
    agileInstance;
    // Key/Name identifier of the State
    _key;
    // Whether the current value differs from the initial value
    isSet = false;
    // Whether the State is a placeholder and only exist in the background
    isPlaceholder = false;
    // First value assigned to the State
    initialStateValue;
    // Current value of the State
    _value;
    // Previous value of the State
    previousStateValue;
    // Next value of the State (which can be used for dynamic State updates)
    nextStateValue;
    // Manages dependencies to other States and subscriptions of UI-Components.
    // It also serves as an interface to the runtime.
    observers = {};
    // Registered side effects of changing the State value
    sideEffects = {};
    /**
     * A State manages a piece of Information
     * that we need to remember globally at a later point in time.
     * While providing a toolkit to use and mutate this piece of Information.
     *
     * You can create as many global States as you need.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/)
     *
     * @public
     * @param agileInstance - Instance of Agile the State belongs to.
     * @param initialValue - Initial value of the State.
     * @param config - Configuration object
     */
    constructor(agileInstance, initialValue, config = {}) {
        config = utils.defineConfig(config, {
            dependents: [],
            isPlaceholder: false,
        });
        this.agileInstance = () => agileInstance;
        this._key = config.key;
        this.observers['value'] = new StateObserver(this, {
            key: config.key,
            dependents: config.dependents,
        });
        this.initialStateValue = utils.copy(initialValue);
        this._value = utils.copy(initialValue);
        this.previousStateValue = utils.copy(initialValue);
        this.nextStateValue = utils.copy(initialValue);
        this.isPlaceholder = true;
        // Set State value to specified initial value
        if (!config.isPlaceholder)
            this.set(initialValue, { overwrite: true });
    }
    /**
     * Assigns a new value to the State
     * and rerenders all subscribed Components.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/properties#value)
     *
     * @public
     * @param value - New State value.
     */
    set value(value) {
        this.set(value);
    }
    /**
     * Returns a reference-free version of the current State value.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/properties#value)
     *
     * @public
     */
    get value() {
        ComputedTracker.tracked(this.observers['value']);
        return utils.copy(this._value);
    }
    /**
     * Updates the key/name identifier of the State.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/properties#key)
     *
     * @public
     * @param value - New key/name identifier.
     */
    set key(value) {
        this.setKey(value);
    }
    /**
     * Returns the key/name identifier of the State.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/properties#key)
     *
     * @public
     */
    get key() {
        return this._key;
    }
    /**
     * Updates the key/name identifier of the State.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/methods/#setkey)
     *
     * @public
     * @param value - New key/name identifier.
     */
    setKey(value) {
        // Update State key
        this._key = value;
        // Update key of Observers
        for (const observerKey in this.observers)
            this.observers[observerKey]._key = value;
        return this;
    }
    /**
     * Assigns a new value to the State
     * and re-renders all subscribed UI-Components.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/methods/#set)
     *
     * @public
     * @param value - New State value
     * @param config - Configuration object
     */
    set(value, config = {}) {
        config = utils.defineConfig(config, {
            force: false,
        });
        const _value = utils.isFunction(value)
            ? value(utils.copy(this._value))
            : value;
        // Ingest the State with the new value into the runtime
        this.observers['value'].ingestValue(_value, config);
        return this;
    }
    /**
     * Ingests the State without any specified new value into the runtime.
     *
     * Since no new value was defined either the new State value is computed
     * based on a compute method (Computed Class)
     * or the `nextStateValue` is taken as the next State value.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/methods/#ingest)
     *
     * @internal
     * @param config - Configuration object
     */
    ingest(config = {}) {
        this.observers['value'].ingest(config);
        return this;
    }
    /**
     *
     * Registers a `callback` function that is executed in the `runtime`
     * as a side effect of State changes.
     *
     * For example, it is called when the State value changes from 'jeff' to 'hans'.
     *
     * A typical side effect of a State change
     * could be the updating of the external Storage value.
     *
     * @internal
     * @param key - Key/Name identifier of the to register side effect.
     * @param callback - Callback function to be fired on each State value change.
     * @param config - Configuration object
     */
    addSideEffect(key, callback, config = {}) {
        config = utils.defineConfig(config, {
            weight: 10,
        });
        if (!utils.isFunction(callback)) {
            LogCodeManager.log('00:03:01', ['Side Effect Callback', 'function']);
            return this;
        }
        this.sideEffects[key] = {
            callback: callback,
            weight: config.weight,
        };
        return this;
    }
    /**
     * Removes a side effect callback with the specified key/name identifier from the State.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/methods/#removesideeffect)
     *
     * @internal
     * @param key - Key/Name identifier of the side effect callback to be removed.
     */
    removeSideEffect(key) {
        delete this.sideEffects[key];
        return this;
    }
    /**
     * Returns a boolean indicating whether a side effect callback with the specified `key`
     * exists in the State or not.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/methods/#hassideeffect)
     *
     * @internal
     * @param key - Key/Name identifier of the side effect callback to be checked for existence.
     */
    hasSideEffect(key) {
        return !!this.sideEffects[key];
    }
}

class StatePersistent extends Persistent {
    // State the Persistent belongs to
    state;
    static storeValueSideEffectKey = 'rebuildStateStorageValue';
    /**
     * Internal Class for managing the permanent persistence of a State.
     *
     * @internal
     * @param state - State to be persisted.
     * @param config - Configuration object
     */
    constructor(state, config = {}) {
        super(state.agileInstance(), {
            loadValue: false,
        });
        config = utils.defineConfig(config, {
            loadValue: true,
            storageKeys: [],
            defaultStorageKey: null,
        });
        this.state = () => state;
        this.instantiatePersistent({
            key: config.key,
            storageKeys: config.storageKeys,
            defaultStorageKey: config.defaultStorageKey,
        });
        // Load/Store persisted value/s for the first time
        if (this.ready && config.loadValue)
            this.initialLoading();
    }
    /**
     * Loads the persisted value into the State
     * or persists the State value in the corresponding Storage.
     * This behaviour depends on whether the State has been persisted before.
     *
     * @internal
     */
    async initialLoading() {
        super.initialLoading().then(() => {
            this.state().isPersisted = true;
        });
    }
    /**
     * Loads the State from the corresponding Storage
     * and sets up side effects that dynamically update
     * the Storage value when the State changes.
     *
     * @internal
     * @param storageItemKey - Storage key of the to load State Instance.
     * | default = Persistent.key |
     * @return Whether the loading of the persisted State Instance and the setting up of the corresponding side effects was successful.
     */
    async loadPersistedValue(storageItemKey) {
        if (!this.ready)
            return false;
        const _storageItemKey = storageItemKey ?? this._key;
        // Load State value from the default Storage
        const loadedValue = await getSharedStorageManager()?.get(_storageItemKey, this.config.defaultStorageKey);
        if (loadedValue == null)
            return false;
        // Assign loaded value to the State
        this.state().set(loadedValue, {
            storage: false,
            overwrite: true,
        });
        // Setup side effects to keep the Storage value in sync
        // with the current State value
        this.setupSideEffects(_storageItemKey);
        return true;
    }
    /**
     * Persists the State in the corresponding Storage
     * and sets up side effects that dynamically update
     * the Storage value when the State changes.
     *
     * @internal
     * @param storageItemKey - Storage key of the to persist State Instance.
     * | default = Persistent.key |
     * @return Whether the persisting of the State Instance and setting up of the corresponding side effects was successful.
     */
    async persistValue(storageItemKey) {
        if (!this.ready)
            return false;
        const _storageItemKey = storageItemKey ?? this._key;
        // Setup side effects to keep the Storage value in sync
        // with the State value
        this.setupSideEffects(_storageItemKey);
        // Initial rebuild Storage for persisting State value in the corresponding Storage
        this.rebuildStorageSideEffect(this.state(), _storageItemKey);
        this.isPersisted = true;
        return true;
    }
    /**
     * Sets up side effects to keep the Storage value in sync
     * with the current State value.
     *
     * @internal
     * @param storageItemKey - Storage key of the persisted State Instance.
     * | default = Persistent.key |
     */
    setupSideEffects(storageItemKey) {
        const _storageItemKey = storageItemKey ?? this._key;
        this.state().addSideEffect(StatePersistent.storeValueSideEffectKey, (instance, config) => {
            this.rebuildStorageSideEffect(this.state(), _storageItemKey, config);
        }, { weight: 0 });
    }
    /**
     * Removes the State from the corresponding Storage.
     * -> State is no longer persisted
     *
     * @internal
     * @param storageItemKey - Storage key of the to remove State Instance.
     * | default = Persistent.key |
     * @return Whether the removal of the persisted State Instance was successful.
     */
    async removePersistedValue(storageItemKey) {
        if (!this.ready)
            return false;
        const _storageItemKey = storageItemKey || this._key;
        this.state().removeSideEffect(StatePersistent.storeValueSideEffectKey);
        getSharedStorageManager()?.remove(_storageItemKey, this.storageKeys);
        this.isPersisted = false;
        return true;
    }
    /**
     * Formats the specified key so that it can be used as a valid Storage key
     * and returns the formatted variant of it.
     *
     * If no formatable key (`undefined`/`null`) was provided,
     * an attempt is made to use the State identifier key as Storage key.
     *
     * @internal
     * @param key - Storage key to be formatted.
     */
    formatKey(key) {
        if (key == null && this.state()._key)
            return this.state()._key;
        if (key == null)
            return;
        if (this.state()._key == null)
            this.state()._key = key;
        return key;
    }
    /**
     * Rebuilds Storage value based on the current State value.
     *
     * @internal
     * @param state - State whose current value to be applied to the Storage value.
     * @param storageItemKey - Storage key of the persisted State Instance.
     * | default = Persistent.key |
     * @param config - Configuration object
     */
    rebuildStorageSideEffect(state, storageItemKey, config = {}) {
        if (config['storage'] == null || config.storage) {
            getSharedStorageManager()?.set(storageItemKey, this.state().getPersistableValue(), this.storageKeys);
        }
    }
}

class EnhancedState extends State {
    // Whether the State is persisted in an external Storage
    isPersisted = false;
    // Manages the permanent persistent in external Storages
    persistent;
    // Method for dynamically computing the State value
    computeValueMethod;
    // Method for dynamically computing the existence of the State
    computeExistsMethod;
    // When an interval is active, the 'intervalId' to clear the interval is temporary stored here
    currentInterval;
    /**
     * An enhanced State manages, like a normal State, a piece of Information
     * that we need to remember globally at a later point in time.
     * While providing a toolkit to use and mutate this piece of Information.
     *
     * The main difference to a normal State is however
     * that an enhanced State provides a wider variety of inbuilt utilities (like a persist, undo, watch functionality)
     * but requires a larger bundle size in return.
     *
     * You can create as many global enhanced States as you need.
     *
     * [Learn more..](https://agile-ts.org/docs/core/agile-instance/methods#createstate)
     *
     * @public
     * @param agileInstance - Instance of Agile the State belongs to.
     * @param initialValue - Initial value of the State.
     * @param config - Configuration object
     */
    constructor(agileInstance, initialValue, config = {}) {
        super(agileInstance, initialValue, config);
        this.computeExistsMethod = (v) => {
            return v != null;
        };
    }
    setKey(value) {
        const oldKey = this._key;
        // Update State key
        super.setKey(value);
        // Update key in Persistent (only if oldKey is equal to persistentKey
        // because otherwise the persistentKey is detached from the State key
        // -> not managed by State anymore)
        if (value != null && this.persistent?._key === oldKey)
            this.persistent?.setKey(value);
        return this;
    }
    /**
     * Undoes the latest State value change.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/methods/#undo)
     *
     * @public
     * @param config - Configuration object
     */
    undo(config = {}) {
        this.set(this.previousStateValue, config);
        return this;
    }
    /**
     * Resets the State value to its initial value.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/methods/#reset)
     *
     * @public
     * @param config - Configuration object
     */
    reset(config = {}) {
        this.set(this.initialStateValue, config);
        return this;
    }
    /**
     * Merges the specified `targetWithChanges` object into the current State value.
     * This merge can differ for different value combinations:
     * - If the current State value is an `object`, it does a partial update for the object.
     * - If the current State value is an `array` and the specified argument is an array too,
     *   it concatenates the current State value with the value of the argument.
     * - If the current State value is neither an `object` nor an `array`, the patch can't be performed.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/methods/#patch)
     *
     * @public
     * @param targetWithChanges - Object to be merged into the current State value.
     * @param config - Configuration object
     */
    patch(targetWithChanges, config = {}) {
        config = utils.defineConfig(config, {
            addNewProperties: true,
        });
        // Check if the given conditions are suitable for a patch action
        if (!utils.isValidObject(this.nextStateValue, true)) {
            LogCodeManager.log('14:03:02');
            return this;
        }
        if (!utils.isValidObject(targetWithChanges, true)) {
            LogCodeManager.log('00:03:01', ['TargetWithChanges', 'object']);
            return this;
        }
        // Merge targetWithChanges object into the nextStateValue
        if (Array.isArray(targetWithChanges) &&
            Array.isArray(this.nextStateValue)) {
            this.nextStateValue = [
                ...this.nextStateValue,
                ...targetWithChanges,
            ];
        }
        else {
            this.nextStateValue = utils.flatMerge(this.nextStateValue, targetWithChanges, { addNewProperties: config.addNewProperties });
        }
        // Ingest updated 'nextStateValue' into runtime
        this.ingest(utils.removeProperties(config, ['addNewProperties']));
        return this;
    }
    watch(keyOrCallback, callback) {
        const generateKey = utils.isFunction(keyOrCallback);
        let _callback;
        let key;
        if (generateKey) {
            key = utils.generateId();
            _callback = keyOrCallback;
        }
        else {
            key = keyOrCallback;
            _callback = callback;
        }
        if (!utils.isFunction(_callback)) {
            LogCodeManager.log('00:03:01', ['Watcher Callback', 'function']);
            return this;
        }
        this.addSideEffect(key, (instance) => {
            _callback(instance.value, key);
        }, { weight: 0 });
        return generateKey ? key : this;
    }
    /**
     * Removes a watcher callback with the specified key/name identifier from the State.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/methods/#removewatcher)
     *
     * @public
     * @param key - Key/Name identifier of the watcher callback to be removed.
     */
    removeWatcher(key) {
        this.removeSideEffect(key);
        return this;
    }
    /**
     * Fires on the initial State value assignment and then destroys itself.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/methods/#oninaugurated)
     *
     * @public
     * @param callback - A function to be executed after the first State value assignment.
     */
    onInaugurated(callback) {
        const watcherKey = 'InauguratedWatcherKey';
        this.watch(watcherKey, (value, key) => {
            callback(value, key);
            this.removeSideEffect(watcherKey);
        });
        return this;
    }
    /**
     * Repeatedly calls the specified callback function,
     * with a fixed time delay between each call.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/methods/#interval)
     *
     * @public
     * @param handler - A function to be executed every delay milliseconds.
     * @param delay - The time, in milliseconds (thousandths of a second),
     * the timer should delay in between executions of the specified function.
     */
    interval(handler, delay) {
        if (!utils.isFunction(handler)) {
            LogCodeManager.log('00:03:01', ['Interval Callback', 'function']);
            return this;
        }
        if (this.currentInterval) {
            LogCodeManager.log('14:03:03', [], this.currentInterval);
            return this;
        }
        this.currentInterval = setInterval(() => {
            this.set(handler(this._value));
        }, delay ?? 1000);
        return this;
    }
    /**
     * Cancels a active timed, repeating action
     * which was previously established by a call to `interval()`.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/methods/#clearinterval)
     *
     * @public
     */
    clearInterval() {
        if (this.currentInterval) {
            clearInterval(this.currentInterval);
            delete this.currentInterval;
        }
    }
    /**
     * Returns a boolean indicating whether the State exists.
     *
     * It calculates the value based on the `computeExistsMethod()`
     * and whether the State is a placeholder.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/methods/#exists)
     *
     * @public
     */
    get exists() {
        return !this.isPlaceholder && this.computeExistsMethod(this.value);
    }
    /**
     * Defines the method used to compute the existence of the State.
     *
     * It is retrieved on each `exists()` method call
     * to determine whether the State exists or not.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/methods/#computeexists)
     *
     * @public
     * @param method - Method to compute the existence of the State.
     */
    computeExists(method) {
        if (!utils.isFunction(method)) {
            LogCodeManager.log('00:03:01', ['Compute Exists Method', 'function']);
            return this;
        }
        this.computeExistsMethod = method;
        return this;
    }
    /**
     * Defines the method used to compute the value of the State.
     *
     * It is retrieved on each State value change,
     * in order to compute the new State value
     * based on the specified compute method.
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/methods/#computevalue)
     *
     * @public
     * @param method - Method to compute the value of the State.
     */
    computeValue(method) {
        if (!utils.isFunction(method)) {
            LogCodeManager.log('00:03:01', ['Compute Value Method', 'function']);
            return this;
        }
        this.computeValueMethod = method;
        // Initial compute
        // (not directly computing it here since it is computed once in the runtime!)
        this.set(this.nextStateValue);
        return this;
    }
    /**
     * Returns a boolean indicating whether the specified value is equal to the current State value.
     *
     * Equivalent to `===` with the difference that it looks at the value
     * and not on the reference in the case of objects.
     *
     * @public
     * @param value - Value to be compared with the current State value.
     */
    is(value) {
        return utils.equal(value, this.value);
    }
    /**
     * Returns a boolean indicating whether the specified value is not equal to the current State value.
     *
     * Equivalent to `!==` with the difference that it looks at the value
     * and not on the reference in the case of objects.
     *
     * @public
     * @param value - Value to be compared with the current State value.
     */
    isNot(value) {
        return utils.notEqual(value, this.value);
    }
    /**
     * Inverts the current State value.
     *
     * Some examples are:
     * - `'jeff'` -> `'ffej'`
     * - `true` -> `false`
     * - `[1, 2, 3]` -> `[3, 2, 1]`
     * - `10` -> `-10`
     *
     * @public
     */
    invert() {
        switch (typeof this.nextStateValue) {
            case 'boolean':
                this.set(!this.nextStateValue);
                break;
            case 'object':
                if (Array.isArray(this.nextStateValue))
                    this.set(this.nextStateValue.reverse());
                break;
            case 'string':
                this.set(this.nextStateValue.split('').reverse().join(''));
                break;
            case 'number':
                this.set((this.nextStateValue * -1));
                break;
            default:
                LogCodeManager.log('14:03:04', [typeof this.nextStateValue]);
        }
        return this;
    }
    persist(keyOrConfig = {}, config = {}) {
        let _config;
        let key;
        if (utils.isValidObject(keyOrConfig)) {
            _config = keyOrConfig;
            key = this._key;
        }
        else {
            _config = config || {};
            key = keyOrConfig;
        }
        _config = utils.defineConfig(_config, {
            loadValue: true,
            storageKeys: [],
            defaultStorageKey: null,
        });
        // Check if State is already persisted
        if (this.persistent != null && this.isPersisted)
            return this;
        // Create Persistent (-> persist value)
        this.persistent = new StatePersistent(this, {
            loadValue: _config.loadValue,
            storageKeys: _config.storageKeys,
            key: key,
            defaultStorageKey: _config.defaultStorageKey,
        });
        return this;
    }
    /**
     * Fires immediately after the persisted `value`
     * is loaded into the State from a corresponding external Storage.
     *
     * Registering such callback function makes only sense
     * when the State is [persisted](https://agile-ts.org/docs/core/state/methods/#persist).
     *
     * [Learn more..](https://agile-ts.org/docs/core/state/methods/#onload)
     *
     * @public
     * @param callback - A function to be executed after the externally persisted `value` was loaded into the State.
     */
    onLoad(callback) {
        if (!this.persistent)
            return this;
        if (!utils.isFunction(callback)) {
            LogCodeManager.log('00:03:01', ['OnLoad Callback', 'function']);
            return this;
        }
        // Register specified callback
        this.persistent.onLoad = callback;
        // If State is already persisted ('isPersisted') fire specified callback immediately
        if (this.isPersisted)
            callback(true);
        return this;
    }
    /**
     * Returns the persistable value of the State.
     *
     * @internal
     */
    getPersistableValue() {
        return this._value;
    }
}

/**
 * Returns a newly created enhanced State.
 *
 * An enhanced State manages, like a normal State, a piece of Information
 * that we need to remember globally at a later point in time.
 * While providing a toolkit to use and mutate this piece of Information.
 *
 * The main difference to a normal State is however
 * that an enhanced State provides a wider variety of inbuilt utilities (like a persist, undo, watch functionality)
 * but requires a larger bundle size in return.
 *
 * You can create as many global enhanced States as you need.
 *
 * [Learn more..](https://agile-ts.org/docs/core/agile-instance/methods#createstate)
 *
 * @public
 * @param initialValue - Initial value of the State.
 * @param config - Configuration object
 */
function createEnhancedState(initialValue, config = {}) {
    config = utils.defineConfig(config, {
        agileInstance: exports.sharedAgileInstance,
    });
    return new EnhancedState(config.agileInstance, initialValue, utils.removeProperties(config, ['agileInstance']));
}

/**
 * Returns a newly created enhanced State.
 *
 * An enhanced State manages, like a normal State, a piece of Information
 * that we need to remember globally at a later point in time.
 * While providing a toolkit to use and mutate this piece of Information.
 *
 * The main difference to a normal State is however
 * that an enhanced State provides a wider variety of inbuilt utilities (like a persist, undo, watch functionality)
 * but requires a larger bundle size in return.
 *
 * You can create as many global enhanced States as you need.
 *
 * [Learn more..](https://agile-ts.org/docs/core/agile-instance/methods#createstate)
 *
 * @public
 * @param initialValue - Initial value of the State.
 * @param config - Configuration object
 */
function createState(initialValue, config = {}) {
    return createEnhancedState(initialValue, config);
}

/**
 * Returns a newly created State.
 *
 * A State manages a piece of Information
 * that we need to remember globally at a later point in time.
 * While providing a toolkit to use and mutate this piece of Information.
 *
 * You can create as many global States as you need.
 *
 * [Learn more..](https://agile-ts.org/docs/core/agile-instance/methods#createstate)
 *
 * @public
 * @param initialValue - Initial value of the State.
 * @param config - Configuration object
 */
function createLightState(initialValue, config = {}) {
    config = utils.defineConfig(config, {
        agileInstance: exports.sharedAgileInstance,
    });
    return new State(config.agileInstance, initialValue, utils.removeProperties(config, ['agileInstance']));
}

class Computed extends State {
    config;
    // Function to compute the Computed Class value
    computeFunction;
    // All dependencies the Computed Class depends on (including hardCoded and automatically detected dependencies)
    deps = new Set();
    // Only hardCoded dependencies the Computed Class depends on
    hardCodedDeps = [];
    // Helper property to check whether an unknown instance is a Computed,
    // without importing the Computed itself for using 'instanceof' (Treeshaking support)
    isComputed = true;
    /**
     * A Computed is an extension of the State Class
     * that computes its value based on a specified compute function.
     *
     * The computed value will be cached to avoid unnecessary recomputes
     * and is only recomputed when one of its direct dependencies changes.
     *
     * Direct dependencies can be States and Collections.
     * So when, for example, a dependent State value changes, the computed value is recomputed.
     *
     * [Learn more..](https://agile-ts.org/docs/core/computed/)
     *
     * @public
     * @param agileInstance - Instance of Agile the Computed belongs to.
     * @param computeFunction - Function to compute the computed value.
     * @param config - Configuration object
     */
    constructor(agileInstance, computeFunction, config = {}) {
        super(agileInstance, null, {
            key: config.key,
            dependents: config.dependents,
        });
        config = utils.defineConfig(config, {
            computedDeps: [],
            autodetect: !utils.isAsyncFunction(computeFunction),
        });
        this.agileInstance = () => agileInstance;
        this.computeFunction = computeFunction;
        this.config = {
            autodetect: config.autodetect,
        };
        // Extract Observer of passed hardcoded dependency instances
        this.hardCodedDeps = extractRelevantObservers(config.computedDeps).filter((dep) => dep !== undefined);
        this.deps = new Set(this.hardCodedDeps);
        // Make this Observer depend on the specified hard coded dep Observers
        this.deps.forEach((observer) => {
            observer.addDependent(this.observers['value']);
        });
        // Initial recompute to assign the computed initial value to the Computed
        // and autodetect missing dependencies
        this.recompute({ autodetect: config.autodetect, overwrite: true });
    }
    /**
     * Forces a recomputation of the cached value with the compute function.
     *
     * [Learn more..](https://agile-ts.org/docs/core/computed/methods/#recompute)
     *
     * @public
     * @param config - Configuration object
     */
    recompute(config = {}) {
        config = utils.defineConfig(config, {
            autodetect: false,
        });
        this.compute({ autodetect: config.autodetect }).then((result) => {
            this.observers['value'].ingestValue(result, utils.removeProperties(config, ['autodetect']));
        });
        return this;
    }
    /**
     * Assigns a new function to the Computed Class for computing its value.
     *
     * The dependencies of the new compute function are automatically detected
     * and accordingly updated.
     *
     * An initial computation is performed with the new function
     * to change the obsolete cached value.
     *
     * [Learn more..](https://agile-ts.org/docs/core/computed/methods/#updatecomputefunction)
     *
     * @public
     * @param computeFunction - New function to compute the value of the Computed Class.
     * @param deps - Hard coded dependencies on which the Computed Class depends.
     * @param config - Configuration object
     */
    updateComputeFunction(computeFunction, deps = [], config = {}) {
        config = utils.defineConfig(config, {
            autodetect: this.config.autodetect,
        });
        // Make this Observer no longer depend on the old dep Observers
        this.deps.forEach((observer) => {
            observer.removeDependent(this.observers['value']);
        });
        // Update dependencies of Computed
        this.hardCodedDeps = extractRelevantObservers(deps).filter((dep) => dep !== undefined);
        this.deps = new Set(this.hardCodedDeps);
        // Make this Observer depend on the new hard coded dep Observers
        this.deps.forEach((observer) => {
            observer.addDependent(this.observers['value']);
        });
        // Update computeFunction
        this.computeFunction = computeFunction;
        // Recompute to assign the new computed value to the Computed
        // and autodetect missing dependencies
        this.recompute(utils.removeProperties(config, ['overwriteDeps']));
        return this;
    }
    /**
     * Computes and returns the new value of the Computed Class
     * and autodetects used dependencies in the compute function.
     *
     * @internal
     * @param config - Configuration object
     */
    async compute(config = {}) {
        config = utils.defineConfig(config, {
            autodetect: this.config.autodetect,
        });
        // Start auto tracking of Observers on which the computeFunction might depend
        if (config.autodetect)
            ComputedTracker.track();
        const computedValue = this.computeFunction();
        // Handle auto tracked Observers
        if (config.autodetect) {
            const foundDeps = ComputedTracker.getTrackedObservers();
            // Clean up old dependencies
            this.deps.forEach((observer) => {
                if (!foundDeps.includes(observer) &&
                    !this.hardCodedDeps.includes(observer)) {
                    this.deps.delete(observer);
                    observer.removeDependent(this.observers['value']);
                }
            });
            // Make this Observer depend on the newly found dep Observers
            foundDeps.forEach((observer) => {
                if (!this.deps.has(observer)) {
                    this.deps.add(observer);
                    observer.addDependent(this.observers['value']);
                }
            });
        }
        return computedValue;
    }
}

function createComputed(computeFunction, configOrDeps) {
    let _config = {};
    if (Array.isArray(configOrDeps)) {
        _config = utils.defineConfig(_config, {
            computedDeps: configOrDeps,
        });
    }
    else {
        if (configOrDeps)
            _config = configOrDeps;
    }
    _config = utils.defineConfig(_config, { agileInstance: exports.sharedAgileInstance });
    return new Computed(_config.agileInstance, computeFunction, utils.removeProperties(_config, ['agileInstance']));
}

class CollectionPersistent extends Persistent {
    // Collection the Persistent belongs to
    collection;
    static defaultGroupSideEffectKey = 'rebuildGroupStorageValue';
    static storageItemKeyPattern = '_${collectionKey}_item_${itemKey}';
    static storageGroupKeyPattern = '_${collectionKey}_group_${groupKey}';
    /**
     * Internal Class for managing the permanent persistence of a Collection.
     *
     * @internal
     * @param collection - Collection to be persisted.
     * @param config - Configuration object
     */
    constructor(collection, config = {}) {
        super(collection.agileInstance(), {
            loadValue: false,
        });
        config = utils.defineConfig(config, {
            loadValue: true,
            storageKeys: [],
            defaultStorageKey: null,
        });
        this.collection = () => collection;
        this.instantiatePersistent({
            key: config.key,
            storageKeys: config.storageKeys,
            defaultStorageKey: config.defaultStorageKey,
        });
        // Load/Store persisted value/s for the first time
        if (this.ready && config.loadValue)
            this.initialLoading();
    }
    /**
     * Loads the persisted value into the Collection
     * or persists the Collection value in the corresponding Storage.
     * This behaviour depends on whether the Collection has been persisted before.
     *
     * @internal
     */
    async initialLoading() {
        super.initialLoading().then(() => {
            this.collection().isPersisted = true;
        });
    }
    /**
     * Loads Collection Instances (like Items or Groups) from the corresponding Storage
     * and sets up side effects that dynamically update
     * the Storage value when the Collection (Instances) changes.
     *
     * @internal
     * @param storageItemKey - Prefix Storage key of the to load Collection Instances.
     * | default = Persistent.key |
     * @return Whether the loading of the persisted Collection Instances and setting up of the corresponding side effects was successful.
     */
    async loadPersistedValue(storageItemKey) {
        if (!this.ready)
            return false;
        const _storageItemKey = storageItemKey ?? this._key;
        // Check if Collection is already persisted
        // (indicated by the persistence of 'true' at '_storageItemKey')
        const isPersisted = await getSharedStorageManager()?.get(_storageItemKey, this.config.defaultStorageKey);
        // Return 'false' if Collection isn't persisted yet
        if (!isPersisted)
            return false;
        // Helper function to load persisted values into the Collection
        const loadValuesIntoCollection = async () => {
            const defaultGroup = this.collection().getDefaultGroup();
            if (defaultGroup == null)
                return false;
            const defaultGroupStorageKey = CollectionPersistent.getGroupStorageKey(defaultGroup._key, _storageItemKey);
            // Persist default Group and load its value manually to be 100% sure
            // that it was loaded completely
            defaultGroup.loadedInitialValue = false;
            defaultGroup.persist(defaultGroupStorageKey, {
                loadValue: false,
                defaultStorageKey: this.config.defaultStorageKey || undefined,
                storageKeys: this.storageKeys,
                followCollectionPersistKeyPattern: false, // Because of the dynamic 'storageItemKey', the key is already formatted above
            });
            if (defaultGroup.persistent?.ready)
                await defaultGroup.persistent.initialLoading();
            // Persist Items found in the default Group's value
            for (const itemKey of defaultGroup._value) {
                const item = this.collection().getItem(itemKey);
                const itemStorageKey = CollectionPersistent.getItemStorageKey(itemKey, _storageItemKey);
                // Persist an already present Item and load its value manually to be 100% sure
                // that it was loaded completely
                if (item != null) {
                    item.persist(itemStorageKey, {
                        loadValue: false,
                        defaultStorageKey: this.config.defaultStorageKey || undefined,
                        storageKeys: this.storageKeys,
                        followCollectionPersistKeyPattern: false, // Because of the dynamic 'storageItemKey', the key is already formatted above
                    });
                    if (item?.persistent?.ready)
                        await item.persistent.initialLoading();
                }
                // Persist an already present placeholder Item
                // or create a new placeholder Item to load the value in
                // and load its value manually to be 100% sure
                // that it was loaded completely.
                // After a successful loading assign the now valid Item to the Collection.
                else {
                    const placeholderItem = this.collection().getItemWithReference(itemKey);
                    placeholderItem?.persist(itemStorageKey, {
                        loadValue: false,
                        defaultStorageKey: this.config.defaultStorageKey,
                        storageKeys: this.storageKeys,
                        followCollectionPersistKeyPattern: false, // Because of the dynamic 'storageItemKey', the key is already formatted above
                    });
                    if (placeholderItem?.persistent?.ready) {
                        const loadedPersistedValueIntoItem = await placeholderItem.persistent.loadPersistedValue();
                        // If successfully loaded Item value, assign Item to Collection
                        if (loadedPersistedValueIntoItem) {
                            this.collection().assignItem(placeholderItem, {
                                overwrite: true,
                                rebuildGroups: false, // Not necessary since the Groups that include the to assign Item were already rebuild while assigning the loaded value to the Item via 'loadPersistedValue()'
                            });
                            placeholderItem.isPersisted = true;
                            // Manually increase Collection size,
                            // since these Items already exist in the Collection (because of 'getItemWithReference()')
                            // but were placeholder before the persisted value got loaded
                            // -> Collection size wasn't increased in 'assignItem()'
                            this.collection().size += 1;
                        }
                    }
                }
            }
            defaultGroup.loadedInitialValue = true;
            return true;
        };
        const success = await loadValuesIntoCollection();
        // Setup side effects to keep the Storage value in sync
        // with the Collection (Instances) value
        if (success)
            this.setupSideEffects(_storageItemKey);
        return success;
    }
    /**
     * Persists Collection Instances (like Items or Groups) in the corresponding Storage
     * and sets up side effects that dynamically update
     * the Storage value when the Collection (Instances) changes.
     *
     * @internal
     * @param storageItemKey - Prefix Storage key of the to persist Collection Instances.
     * | default = Persistent.key |
     * @return Whether the persisting of the Collection Instances and the setting up of the corresponding side effects was successful.
     */
    async persistValue(storageItemKey) {
        if (!this.ready)
            return false;
        const _storageItemKey = storageItemKey ?? this._key;
        const defaultGroup = this.collection().getDefaultGroup();
        if (defaultGroup == null)
            return false;
        const defaultGroupStorageKey = CollectionPersistent.getGroupStorageKey(defaultGroup._key, _storageItemKey);
        // Set flag in Storage to indicate that the Collection is persisted
        getSharedStorageManager()?.set(_storageItemKey, true, this.storageKeys);
        // Persist default Group
        defaultGroup.persist(defaultGroupStorageKey, {
            defaultStorageKey: this.config.defaultStorageKey || undefined,
            storageKeys: this.storageKeys,
            followCollectionPersistKeyPattern: false, // Because of the dynamic 'storageItemKey', the key is already formatted above
        });
        // Persist Items found in the default Group's value
        for (const itemKey of defaultGroup._value) {
            const item = this.collection().getItem(itemKey);
            const itemStorageKey = CollectionPersistent.getItemStorageKey(itemKey, _storageItemKey);
            item?.persist(itemStorageKey, {
                defaultStorageKey: this.config.defaultStorageKey || undefined,
                storageKeys: this.storageKeys,
                followCollectionPersistKeyPattern: false, // Because of the dynamic 'storageItemKey', the key is already formatted above
            });
        }
        // Setup side effects to keep the Storage value in sync
        // with the Collection (Instances) value
        this.setupSideEffects(_storageItemKey);
        this.isPersisted = true;
        return true;
    }
    /**
     * Sets up side effects to keep the Storage value in sync
     * with the Collection (Instances) value.
     *
     * @internal
     * @param storageItemKey - Prefix Storage key of the to remove Collection Instances.
     * | default = Persistent.key |
     */
    setupSideEffects(storageItemKey) {
        const _storageItemKey = storageItemKey ?? this._key;
        const defaultGroup = this.collection().getDefaultGroup();
        if (defaultGroup == null)
            return;
        // Add side effect to the default Group
        // that adds and removes Items from the Storage based on the Group value
        defaultGroup.addSideEffect(CollectionPersistent.defaultGroupSideEffectKey, (instance) => this.rebuildStorageSideEffect(instance, _storageItemKey), { weight: 0 });
    }
    /**
     * Removes the Collection from the corresponding Storage.
     * -> Collection is no longer persisted
     *
     * @internal
     * @param storageItemKey - Prefix Storage key of the persisted Collection Instances.
     * | default = Persistent.key |
     * @return Whether the removal of the Collection Instances was successful.
     */
    async removePersistedValue(storageItemKey) {
        if (!this.ready)
            return false;
        const _storageItemKey = storageItemKey ?? this._key;
        const defaultGroup = this.collection().getDefaultGroup();
        if (!defaultGroup)
            return false;
        const defaultGroupStorageKey = CollectionPersistent.getGroupStorageKey(defaultGroup._key, _storageItemKey);
        // Remove Collection is persisted indicator flag from Storage
        getSharedStorageManager()?.remove(_storageItemKey, this.storageKeys);
        // Remove default Group from the Storage
        defaultGroup.persistent?.removePersistedValue(defaultGroupStorageKey);
        defaultGroup.removeSideEffect(CollectionPersistent.defaultGroupSideEffectKey);
        // Remove Items found in the default Group's value from the Storage
        for (const itemKey of defaultGroup._value) {
            const item = this.collection().getItem(itemKey);
            const itemStorageKey = CollectionPersistent.getItemStorageKey(itemKey, _storageItemKey);
            item?.persistent?.removePersistedValue(itemStorageKey);
        }
        this.isPersisted = false;
        return true;
    }
    /**
     * Formats the specified key so that it can be used as a valid Storage key
     * and returns the formatted variant of it.
     *
     * If no formatable key (`undefined`/`null`) was provided,
     * an attempt is made to use the Collection identifier key as Storage key.
     *
     * @internal
     * @param key - Storage key to be formatted.
     */
    formatKey(key) {
        if (key == null && this.collection()._key)
            return this.collection()._key;
        if (key == null)
            return;
        if (this.collection()._key == null)
            this.collection()._key = key;
        return key;
    }
    /**
     * Adds and removes Items from the Storage based on the Group value.
     *
     * @internal
     * @param group - Group whose Items are to be dynamically added or removed from the Storage.
     * @param storageItemKey - Prefix Storage key of the persisted Collection Instances.
     * | default = Persistent.key |
     */
    rebuildStorageSideEffect(group, storageItemKey) {
        const collection = group.collection();
        const _storageItemKey = storageItemKey || collection.persistent?._key;
        // Return if no Item got added or removed
        // because then the changed Item performs the Storage update itself
        if (group.previousStateValue.length === group._value.length)
            return;
        // Extract Item keys that got added or removed from the Group
        const addedKeys = group._value.filter((key) => !group.previousStateValue.includes(key));
        const removedKeys = group.previousStateValue.filter((key) => !group._value.includes(key));
        // Persist newly added Items
        addedKeys.forEach((itemKey) => {
            const item = collection.getItem(itemKey);
            const itemStorageKey = CollectionPersistent.getItemStorageKey(itemKey, _storageItemKey);
            if (item != null && !item.isPersisted)
                item.persist(itemStorageKey, {
                    defaultStorageKey: this.config.defaultStorageKey || undefined,
                    storageKeys: this.storageKeys,
                    followCollectionPersistKeyPattern: false, // Because of the dynamic 'storageItemKey', the key is already formatted above
                });
        });
        // Remove removed Items from the Storage
        removedKeys.forEach((itemKey) => {
            const item = collection.getItem(itemKey);
            const itemStorageKey = CollectionPersistent.getItemStorageKey(itemKey, _storageItemKey);
            if (item != null && item.isPersisted)
                item.persistent?.removePersistedValue(itemStorageKey);
        });
    }
    /**
     * Builds valid Item Storage key based on the 'Collection Item Persist Pattern'.
     *
     * @internal
     * @param itemKey - Key identifier of Item
     * @param collectionKey - Key identifier of Collection
     */
    static getItemStorageKey(itemKey, collectionKey) {
        if (itemKey == null || collectionKey == null)
            LogCodeManager.log('1A:02:00');
        if (itemKey == null)
            itemKey = 'unknown';
        if (collectionKey == null)
            collectionKey = 'unknown';
        return this.storageItemKeyPattern
            .replace('${collectionKey}', collectionKey.toString())
            .replace('${itemKey}', itemKey.toString());
    }
    /**
     * Builds valid Item Storage key based on the 'Collection Group Persist Pattern'.
     *
     * @internal
     * @param groupKey - Key identifier of Group
     * @param collectionKey - Key identifier of Collection
     */
    static getGroupStorageKey(groupKey, collectionKey) {
        if (groupKey == null || collectionKey == null)
            LogCodeManager.log('1A:02:01');
        if (groupKey == null)
            groupKey = 'unknown';
        if (collectionKey == null)
            collectionKey = 'unknown';
        return this.storageGroupKeyPattern
            .replace('${collectionKey}', collectionKey.toString())
            .replace('${groupKey}', groupKey.toString());
    }
}

class Item extends EnhancedState {
    // Collection the Group belongs to
    collection;
    static updateGroupSideEffectKey = 'rebuildGroup';
    // Key/Name identifiers of Selectors which have selected the Item
    selectedBy = new Set();
    /**
     * An extension of the State Class that represents a single data object of a Collection.
     *
     * It can be used independently, but is always synchronized with the Collection.
     *
     * @public
     * @param collection - Collection to which the Item belongs.
     * @param data - Data object to be represented by the Item.
     * @param config - Configuration object
     */
    constructor(collection, data, config = {}) {
        super(collection.agileInstance(), data, {
            isPlaceholder: config.isPlaceholder,
            key: data[collection.config.primaryKey], // Set key/name of Item to identifier at primaryKey property
        });
        this.collection = () => collection;
        // Add side effect to Item
        // that rebuilds all Groups containing the Item whenever it changes
        if (this._key != null) {
            this.addRebuildGroupThatIncludeItemKeySideEffect(this._key);
        }
    }
    /**
     * Updates the key/name identifier of Item.
     *
     * @internal
     * @param value - New key/name identifier.
     * @param config - Configuration object
     */
    setKey(value, config = {}) {
        super.setKey(value);
        config = utils.defineConfig(config, {
            sideEffects: {
                enabled: true,
                exclude: [],
            },
            background: false,
            force: false,
            storage: true,
            overwrite: false,
        });
        if (value == null)
            return this;
        // Update 'rebuildGroupsThatIncludeItemKey' side effect to the new itemKey
        this.removeSideEffect(Item.updateGroupSideEffectKey);
        this.addRebuildGroupThatIncludeItemKeySideEffect(value);
        // Update itemKey in Item value
        // (After updating the side effect, because otherwise it would call the old side effect)
        this.patch({ [this.collection().config.primaryKey]: value }, config);
        return this;
    }
    persist(keyOrConfig = {}, config = {}) {
        let _config;
        let key;
        if (utils.isValidObject(keyOrConfig)) {
            _config = keyOrConfig;
            key = this._key;
        }
        else {
            _config = config || {};
            key = keyOrConfig;
        }
        _config = utils.defineConfig(_config, {
            loadValue: true,
            followCollectionPersistKeyPattern: true,
            storageKeys: [],
            defaultStorageKey: null,
        });
        // Create storageItemKey based on Collection key/name identifier
        if (_config.followCollectionPersistKeyPattern) {
            key = CollectionPersistent.getItemStorageKey(key || this._key, this.collection()._key);
        }
        // Persist Item
        super.persist(key, {
            loadValue: _config.loadValue,
            storageKeys: _config.storageKeys,
            defaultStorageKey: _config.defaultStorageKey,
        });
        return this;
    }
    /**
     * Adds side effect to Item
     * that rebuilds all Groups containing the specified Item identifier
     * whenever the Item changes.
     *
     * @internal
     * @param itemKey - Item identifier that has to be contained in Groups.
     */
    addRebuildGroupThatIncludeItemKeySideEffect(itemKey) {
        this.addSideEffect(Item.updateGroupSideEffectKey, (instance, config) => {
            instance.collection().rebuildGroupsThatIncludeItemKey(itemKey, config);
        }, { weight: 100 });
    }
}

class Selector extends EnhancedState {
    // Collection the Selector belongs to
    collection;
    static rebuildSelectorSideEffectKey = 'rebuildSelector';
    static rebuildItemSideEffectKey = 'rebuildItem';
    // Item the Selector represents
    _item;
    // Key/Name identifier of the Item the Selector represents
    _itemKey;
    /**
     * A Selector represents an Item from a Collection in the long term.
     * It can be mutated dynamically and remains in sync with the Collection.
     *
     * Components that need one piece of data from a Collection such as the "current user"
     * would benefit from using Selectors.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/selector)
     *
     * @public
     * @param collection - Collection to which the Selector belongs.
     * @param itemKey - Key/Name identifier of the Item to be represented by the Selector.
     * @param config - Configuration object
     */
    constructor(collection, itemKey, config = {}) {
        config = utils.defineConfig(config, {
            isPlaceholder: false,
        });
        super(collection.agileInstance(), null, config);
        this.collection = () => collection;
        this._item = null;
        this._itemKey = !config.isPlaceholder && itemKey != null ? itemKey : null;
        this._key = config?.key;
        this.isPlaceholder = true; // Because hasn't selected any Item yet
        // Initial select of the Item
        if (this._itemKey != null)
            this.select(itemKey, { overwrite: true });
    }
    /**
     * Returns the `itemKey` currently selected by the Selector.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/selector/properties#itemkey)
     *
     * @public
     */
    get itemKey() {
        return this._itemKey;
    }
    /**
     * Updates the currently selected Item of the Selector
     * based on the specified `itemKey`.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/selector/properties#itemkey)
     *
     * @public
     * @param value - New key/name identifier of the Item to be represented by the Selector.
     */
    set itemKey(value) {
        this.select(value);
    }
    /**
     * Retrieves the Item currently selected by the Selector.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/selector/properties#item)
     *
     * @public
     */
    get item() {
        return this._item;
    }
    /**
     * Updates the currently selected Item of the Selector
     * based on the specified Item.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/selector/properties#item)
     *
     * @public
     * @param value - New Item to be represented by the Selector.
     */
    set item(value) {
        if (value?._key)
            this.select(value._key);
    }
    /**
     * Updates the currently selected Item of the Selector.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/selector/methods#select)
     *
     * @public
     * @param itemKey - New key/name identifier of the Item to be represented by the Selector.
     * @param config - Configuration object
     */
    select(itemKey, config = {}) {
        config = utils.defineConfig(config, {
            background: false,
            sideEffects: {
                enabled: true,
                exclude: [],
            },
            force: false,
            overwrite: this._item?.isPlaceholder ?? false,
            storage: true,
        });
        // Don't select Item if Collection is not correctly instantiated yet
        // (because only after a successful instantiation the Collection
        // contains the Items which are essential for a proper selection)
        if ((!this.collection().isInstantiated || this.hasSelected(itemKey)) &&
            !config.force)
            return this;
        // Unselect old Item
        this.unselect({ background: itemKey != null });
        if (itemKey == null)
            return this;
        // Retrieve new Item from Collection
        const newItem = this.collection().getItemWithReference(itemKey);
        // Select new Item
        this._itemKey = itemKey;
        this._item = newItem;
        newItem.selectedBy.add(this._key);
        // Add side effect to the newly selected Item
        // that rebuilds the Selector value depending on the current Item value
        newItem.addSideEffect(Selector.rebuildSelectorSideEffectKey, (instance, config) => this.rebuildSelector(config), { weight: 100 });
        // Add side effect to Selector
        // that updates the Item value depending on the current Selector value
        this.addSideEffect(Selector.rebuildItemSideEffectKey, (instance, config) => {
            if (!instance._item?.isPlaceholder)
                instance._item?.set(instance._value, {
                    ...config,
                    ...{
                        sideEffects: {
                            enabled: true,
                            exclude: [Selector.rebuildSelectorSideEffectKey], // Exclude to avoid endless loops
                        },
                    },
                });
        }, { weight: 90 });
        // Rebuild the Selector to properly 'instantiate' the newly selected Item
        this.rebuildSelector(config);
        return this;
    }
    /**
     * Reselects the currently selected Item.
     *
     * This might be helpful if the Selector failed to select the Item correctly before
     * and therefore should try to select it again.
     *
     * You can use the 'hasSelected()' method to check
     * whether the 'selected' Item is selected correctly.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/selector/methods#reselect)
     *
     * @public
     * @param config - Configuration object
     */
    reselect(config = {}) {
        if ((this._itemKey != null && !this.hasSelected(this._itemKey)) ||
            config.force)
            this.select(this._itemKey, config);
        return this;
    }
    /**
     * Unselects the currently selected Item.
     *
     * Therefore, it sets the `itemKey` and `item` property to `undefined`,
     * since the Selector no longer represents any Item.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/selector/methods#unselect)
     *
     * @public
     * @param config - Configuration object
     */
    unselect(config = {}) {
        // Retrieve Item from the Collection because 'this._item' might be outdated
        const item = this.collection().getItem(this._itemKey, {
            notExisting: true,
        });
        // Unselect Item
        if (item != null) {
            item.selectedBy.delete(this._key);
            item.removeSideEffect(Selector.rebuildSelectorSideEffectKey);
            item.removeSideEffect(Selector.rebuildItemSideEffectKey);
            if (item.isPlaceholder && this._itemKey != null)
                delete this.collection().data[this._itemKey];
        }
        // Reset Selector
        this._item = null;
        this._itemKey = null;
        this.rebuildSelector(config);
        this.isPlaceholder = true;
        return this;
    }
    /**
     * Returns a boolean indicating whether an Item with the specified `itemKey`
     * is selected by the Selector or not.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/selector/methods#hasselected)
     *
     * @public
     * @param itemKey - Key/Name identifier of the Item.
     * @param correctlySelected - Whether the Item has to be selected correctly.
     */
    hasSelected(itemKey, correctlySelected = true) {
        if (correctlySelected) {
            return (this._itemKey === itemKey &&
                this._item != null &&
                this._item.selectedBy.has(this._key));
        }
        return this._itemKey === itemKey;
    }
    /**
     * Rebuilds the Selector.
     * During this process, it updates the Selector `value` based on the Item `value`.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/selector/methods#rebuild)
     *
     * @public
     * @param config - Configuration object
     */
    rebuildSelector(config = {}) {
        // Assign 'undefined' to the Selector value if no Item is set
        if (this._item == null || this._item.isPlaceholder) {
            this.set(null, config);
            return this;
        }
        // Assign the current Item value to the Selector value
        this.set(this._item._value, config);
        return this;
    }
}

class GroupObserver extends Observer {
    // Group the Observer belongs to
    group;
    // Next output applied to the Group
    nextGroupOutput;
    /**
     * A Group Observer manages the subscriptions to Subscription Containers (UI-Components)
     * and dependencies to other Observers (Agile Classes)
     * for a Group Class.
     *
     * @internal
     * @param group - Instance of Group the Observer belongs to.
     * @param config - Configuration object
     */
    constructor(group, config = {}) {
        super(group.agileInstance(), utils.defineConfig(config, { value: group._output }));
        this.group = () => group;
        this.nextGroupOutput = utils.copy(group._output);
    }
    /**
     * Rebuilds the Group and passes the Group Observer
     * into the runtime wrapped into a Runtime-Job
     * where it is executed accordingly.
     *
     * During the execution the runtime applies the `nextGroupOutput` to the Group,
     * updates its dependents and re-renders the UI-Components it is subscribed to.
     *
     * @internal
     * @param config - Configuration object
     */
    ingest(config = {}) {
        this.ingestOutput(this.group().nextGroupOutput, config);
    }
    /**
     * Passes the Group Observer into the runtime wrapped into a Runtime-Job
     * where it is executed accordingly.
     *
     * During the execution the runtime applies the specified `nextGroupOutput` to the Group,
     * updates its dependents and re-renders the UI-Components it is subscribed to.
     *
     * @internal
     * @param newGroupOutput - New Group Output to be applied to the Group.
     * @param config - Configuration object.
     */
    ingestOutput(newGroupOutput, config = {}) {
        const group = this.group();
        config = utils.defineConfig(config, {
            perform: true,
            force: false,
        });
        // Force overwriting the Group value if it is a placeholder.
        // After assigning a value to the Group, the Group is supposed to be no placeholder anymore.
        if (group.isPlaceholder) {
            config.force = true;
        }
        // Assign next Group output to Observer
        this.nextGroupOutput = utils.copy(newGroupOutput);
        // Check if current Group output and to assign Group output are equal
        if (utils.equal(group._output, this.nextGroupOutput) && !config.force)
            return;
        // Create Runtime-Job
        const job = new RuntimeJob(this, {
            ...utils.removeProperties(config, ['perform']),
            ...{
                key: config.key ??
                    `${this._key != null ? this._key + '_' : ''}${utils.generateId()}_output`,
            },
        });
        // Pass created Job into the Runtime
        this.agileInstance().runtime.ingest(job, {
            perform: config.perform,
        });
    }
    /**
     * Method executed by the Runtime to perform the Runtime-Job,
     * previously ingested via the `ingest()` or `ingestItems()` method.
     *
     * Thereby the previously defined `nextGroupOutput` is assigned to the Group.
     *
     * @internal
     * @param job - Runtime-Job to be performed.
     */
    perform(job) {
        const observer = job.observer;
        const group = observer.group();
        // Assign new Group output
        group._output = utils.copy(observer.nextGroupOutput);
        group.nextGroupOutput = utils.copy(observer.nextGroupOutput);
        // Assign new public output to the Observer (output used by the Integrations)
        job.observer.previousValue = utils.copy(job.observer.value);
        job.observer.value = utils.copy(group._output);
    }
}

class Group extends EnhancedState {
    // Collection the Group belongs to
    collection;
    static rebuildGroupSideEffectKey = 'rebuildGroup';
    // Item values represented by the Group
    _output = [];
    // Next output of the Group (which can be used for dynamic Group updates)
    nextGroupOutput = [];
    // Precise itemKeys of the Group only include itemKeys
    // that actually exist in the corresponding Collection
    _preciseItemKeys = [];
    // Manages dependencies to other States and subscriptions of UI-Components.
    // It also serves as an interface to the runtime.
    observers = {};
    // Keeps track of all Item identifiers for Items that couldn't be found in the Collection
    notFoundItemKeys = [];
    // Whether the initial value was loaded from the corresponding Persistent
    // https://github.com/agile-ts/agile/issues/155
    loadedInitialValue = true;
    /**
     * An extension of the State Class that categorizes and preserves the ordering of structured data.
     * It allows us to cluster together data from a Collection as an array of Item keys.
     *
     * Note that a Group doesn't store the actual Items. It only keeps track of the Item keys
     * and retrieves the fitting Items when needed.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/group/)
     *
     * @public
     * @param collection - Collection to which the Group belongs.
     * @param initialItems - Key/Name identifiers of the Items to be clustered by the Group.
     * @param config - Configuration object
     */
    constructor(collection, initialItems = [], config = {}) {
        super(collection.agileInstance(), initialItems, config);
        // Have to redefine the value Observer (observers['value']) again,
        // although it was technically set in the State Parent
        // https://github.com/microsoft/TypeScript/issues/1617
        this.observers['value'] = new StateObserver(this, {
            key: config.key,
        });
        this.observers['output'] = new GroupObserver(this, {
            key: config.key,
        });
        this.collection = () => collection;
        // Add side effect to Group
        // that rebuilds the Group whenever the Group value changes
        this.addSideEffect(Group.rebuildGroupSideEffectKey, (state, config) => {
            this.rebuild(config?.any?.trackedChanges || [], config);
        });
        // Initial rebuild
        this.rebuild();
    }
    /**
     * Returns the values of the Items clustered by the Group.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/group/properties#output)
     *
     * @public
     */
    get output() {
        ComputedTracker.tracked(this.observers['output']);
        return utils.copy(this._output);
    }
    set output(value) {
        LogCodeManager.log('1C:03:00', [this._key]);
    }
    /**
     * Returns a boolean indicating whether an Item with the specified `itemKey`
     * is clustered in the Group or not.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/group/methods/#has)
     *
     * @public
     * @param itemKey - Key/Name identifier of the Item.
     */
    has(itemKey) {
        return this.value.indexOf(itemKey) !== -1;
    }
    /**
     * Returns the count of Items clustered by the Group.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/group/properties#size)
     *
     * @public
     */
    get size() {
        return this.value.length;
    }
    /**
     * Removes an Item with the specified key/name identifier from the Group,
     * if it exists in the Group.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/group/methods#remove)
     *
     * @public
     * @param itemKeys - Key/Name identifier/s of the Item/s to be removed.
     * @param config - Configuration object
     */
    remove(itemKeys, config = {}) {
        const _itemKeys = utils.normalizeArray(itemKeys);
        const notExistingItemKeysInCollection = [];
        const notExistingItemKeys = [];
        let newGroupValue = utils.copy(this.nextStateValue);
        // Need to temporary update the preciseItemKeys
        // since in the rebuild one action (trackedChanges) is performed after the other
        // which requires a dynamic updated index
        const updatedPreciseItemKeys = utils.copy(this._preciseItemKeys);
        config = utils.defineConfig(config, {
            softRebuild: true,
            any: {},
        });
        config.any['trackedChanges'] = []; // TODO should be improved since the 'any' property is very vague
        // Remove itemKeys from Group
        _itemKeys.forEach((itemKey) => {
            const exists = newGroupValue.includes(itemKey);
            // Check if itemKey exists in Group
            if (!exists) {
                notExistingItemKeys.push(itemKey);
                notExistingItemKeysInCollection.push(itemKey);
                return;
            }
            // Track changes to soft rebuild the Group when rebuilding the Group in a side effect
            if (config.softRebuild) {
                const index = updatedPreciseItemKeys.findIndex((ik) => ik === itemKey);
                if (index !== -1) {
                    updatedPreciseItemKeys.splice(index, 1);
                    config.any['trackedChanges'].push({
                        index,
                        method: exports.TrackedChangeMethod.REMOVE,
                        key: itemKey,
                    });
                }
            }
            // Check if itemKey exists in Collection
            if (!this.collection().getItem(itemKey))
                notExistingItemKeysInCollection.push(itemKey);
            // Remove itemKey from Group
            newGroupValue = newGroupValue.filter((key) => key !== itemKey);
        });
        // Return if none of the specified itemKeys exists
        if (notExistingItemKeys.length >= _itemKeys.length)
            return this;
        // If all removed itemKeys don't exist in the Collection
        // -> no rerender necessary since the output won't change
        if (notExistingItemKeysInCollection.length >= _itemKeys.length)
            config.background = true;
        this.set(newGroupValue, utils.removeProperties(config, ['softRebuild']));
        return this;
    }
    /**
     * Appends new Item/s to the end of the Group.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/group/methods#add)
     *
     * @public
     * @param itemKeys - Key/Name identifier/s of Item/s to be added.
     * @param config - Configuration object
     */
    add(itemKeys, config = {}) {
        const _itemKeys = utils.normalizeArray(itemKeys);
        const notExistingItemKeysInCollection = [];
        const existingItemKeys = [];
        const newGroupValue = utils.copy(this.nextStateValue);
        // Need to temporary update the preciseItemKeys
        // since in the rebuild one action (trackedChanges) is performed after the other
        // which requires a dynamic updated index
        const updatedPreciseItemKeys = utils.copy(this._preciseItemKeys);
        config = utils.defineConfig(config, {
            method: 'push',
            softRebuild: true,
            any: {},
        });
        config.any['trackedChanges'] = []; // TODO should be improved since the 'any' property is very vague
        // Add itemKeys to Group
        _itemKeys.forEach((itemKey) => {
            const exists = newGroupValue.includes(itemKey);
            // Check if itemKey exists in Collection
            if (!this.collection().getItem(itemKey))
                notExistingItemKeysInCollection.push(itemKey);
            // Handle existing Item
            if (exists) {
                existingItemKeys.push(itemKey);
                return;
            }
            // Track changes to soft rebuild the Group when rebuilding the Group in a side effect
            if (config.softRebuild) {
                const index = config.method === 'push' ? updatedPreciseItemKeys.length : 0;
                updatedPreciseItemKeys.push(itemKey);
                config.any['trackedChanges'].push({
                    index,
                    method: exports.TrackedChangeMethod.ADD,
                    key: itemKey,
                });
            }
            // Add new itemKey to Group
            newGroupValue[config.method || 'push'](itemKey);
        });
        // Return if all specified itemKeys already exist
        if (existingItemKeys.length >= _itemKeys.length)
            return this;
        // If all added itemKeys don't exist in the Collection
        // -> no rerender necessary since the output won't change
        if (notExistingItemKeysInCollection.concat(existingItemKeys).length >=
            _itemKeys.length)
            config.background = true;
        this.set(newGroupValue, utils.removeProperties(config, ['method', 'softRebuild']));
        return this;
    }
    /**
     * Replaces the old `itemKey` with a new specified `itemKey`.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/group/methods#replace)
     *
     * @public
     * @param oldItemKey - Old `itemKey` to be replaced.
     * @param newItemKey - New `itemKey` to replace the before specified old `itemKey`.
     * @param config - Configuration object
     */
    replace(oldItemKey, newItemKey, config = {}) {
        const newGroupValue = utils.copy(this._value);
        newGroupValue.splice(newGroupValue.indexOf(oldItemKey), 1, newItemKey);
        this.set(newGroupValue, config);
        return this;
    }
    /**
     * Retrieves all existing Items of the Group from the corresponding Collection and returns them.
     * Items that aren't present in the Collection are skipped.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/group/methods#getitems)
     *
     * @public
     */
    getItems() {
        return this.value
            .map((itemKey) => this.collection().getItem(itemKey))
            .filter((item) => item !== undefined);
    }
    persist(keyOrConfig = {}, config = {}) {
        let _config;
        let key;
        if (utils.isValidObject(keyOrConfig)) {
            _config = keyOrConfig;
            key = this._key;
        }
        else {
            _config = config || {};
            key = keyOrConfig;
        }
        _config = utils.defineConfig(_config, {
            loadValue: true,
            followCollectionPersistKeyPattern: true,
            storageKeys: [],
            defaultStorageKey: null,
        });
        // Create storageItemKey based on Collection key/name identifier
        if (_config.followCollectionPersistKeyPattern) {
            key = CollectionPersistent.getGroupStorageKey(key || this._key, this.collection()._key);
        }
        // Persist Group
        super.persist(key, {
            loadValue: _config.loadValue,
            storageKeys: _config.storageKeys,
            defaultStorageKey: _config.defaultStorageKey,
        });
        return this;
    }
    /**
     * Rebuilds the output of the Group
     * and ingests it into the runtime.
     *
     * In doing so, it traverses the Group `value` (Item identifiers)
     * and fetches the fitting Items accordingly.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/group/methods#rebuild)
     *
     * @internal
     * @param trackedChanges - Changes that were tracked between two rebuilds.
     * @param config - Configuration object
     */
    rebuild(trackedChanges = [], config = {}) {
        // Don't rebuild Group if Collection isn't correctly instantiated yet
        // (because only after a successful instantiation the Collection
        // contains the Items which are essential for a proper rebuild)
        if (!this.collection().isInstantiated)
            return this;
        // Item keys that couldn't be found in the Collection
        const notFoundItemKeys = [];
        // Soft rebuild the Collection (-> rebuild only parts of the Collection)
        if (trackedChanges.length > 0) {
            trackedChanges.forEach((change) => {
                const item = this.collection().getItem(change.key);
                switch (change.method) {
                    case exports.TrackedChangeMethod.ADD:
                        // this._value.splice(change.index, 0, change.key); // Already updated in 'add' method
                        if (item != null) {
                            this._preciseItemKeys.splice(change.index, 0, change.key);
                            this.nextGroupOutput.splice(change.index, 0, utils.copy(item._value));
                        }
                        else {
                            notFoundItemKeys.push(change.key);
                        }
                        break;
                    case exports.TrackedChangeMethod.UPDATE:
                        if (item != null) {
                            this.nextGroupOutput[change.index] = utils.copy(item._value);
                        }
                        else {
                            notFoundItemKeys.push(change.key);
                        }
                        break;
                    case exports.TrackedChangeMethod.REMOVE:
                        // this._value.splice(change.index, 1); // Already updated in 'remove' method
                        this._preciseItemKeys.splice(change.index, 1);
                        this.nextGroupOutput.splice(change.index, 1);
                        break;
                }
            });
            this.observers['output'].ingest(config);
        }
        // Hard rebuild the whole Collection
        else {
            const groupItemValues = [];
            // Reset precise itemKeys array to rebuild it from scratch
            this._preciseItemKeys = [];
            // Fetch Items from Collection
            this._value.forEach((itemKey) => {
                const item = this.collection().getItem(itemKey);
                if (item != null) {
                    groupItemValues.push(item._value);
                    this._preciseItemKeys.push(itemKey);
                }
                else
                    notFoundItemKeys.push(itemKey);
            });
            // Ingest rebuilt Group output into the Runtime
            this.observers['output'].ingestOutput(groupItemValues, config);
        }
        // Logging
        if (notFoundItemKeys.length > 0 && this.loadedInitialValue) {
            LogCodeManager.log('1C:02:00', [this.collection()._key, this._key], notFoundItemKeys);
        }
        this.notFoundItemKeys = notFoundItemKeys;
        return this;
    }
}
exports.TrackedChangeMethod = void 0;
(function (TrackedChangeMethod) {
    TrackedChangeMethod[TrackedChangeMethod["ADD"] = 0] = "ADD";
    TrackedChangeMethod[TrackedChangeMethod["REMOVE"] = 1] = "REMOVE";
    TrackedChangeMethod[TrackedChangeMethod["UPDATE"] = 2] = "UPDATE";
})(exports.TrackedChangeMethod || (exports.TrackedChangeMethod = {}));

class Collection {
    // Agile Instance the Collection belongs to
    agileInstance;
    config;
    initialConfig;
    // Key/Name identifier of the Collection
    _key;
    // Amount of the Items stored in the Collection
    size = 0;
    // Items stored in the Collection
    data = {};
    // Whether the Collection is persisted in an external Storage
    isPersisted = false;
    // Manages the permanent persistent in external Storages
    persistent;
    // Registered Groups of Collection
    groups = {};
    // Registered Selectors of Collection
    selectors = {};
    // Whether the Collection was instantiated correctly
    isInstantiated = false;
    // Helper property to check whether an unknown instance is a Collection,
    // without importing the Collection itself for using 'instanceof' (Treeshaking support)
    isCollection = true;
    /**
     * A Collection manages a reactive set of Information
     * that we need to remember globally at a later point in time.
     * While providing a toolkit to use and mutate this set of Information.
     *
     * It is designed for arrays of data objects following the same pattern.
     *
     * Each of these data object must have a unique `primaryKey` to be correctly identified later.
     *
     * You can create as many global Collections as you need.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/)
     *
     * @public
     * @param agileInstance - Instance of Agile the Collection belongs to.
     * @param config - Configuration object
     */
    constructor(agileInstance, config = {}) {
        this.agileInstance = () => agileInstance;
        let _config = typeof config === 'function' ? config(this) : config;
        _config = utils.defineConfig(_config, {
            primaryKey: 'id',
            groups: {},
            selectors: {},
            defaultGroupKey: 'default',
        });
        this._key = _config.key;
        this.config = {
            defaultGroupKey: _config.defaultGroupKey,
            primaryKey: _config.primaryKey,
        };
        this.initialConfig = _config;
        this.initGroups(_config.groups);
        this.initSelectors(_config.selectors);
        this.isInstantiated = true;
        // Add 'initialData' to Collection
        // (after 'isInstantiated' to add them properly to the Collection)
        if (_config.initialData)
            this.collect(_config.initialData);
        // Reselect Selector Items
        // Necessary because the selection of an Item
        // hasn't worked with a not correctly 'instantiated' Collection before
        for (const key in this.selectors)
            this.selectors[key].reselect();
        // Rebuild of Groups
        // Not necessary because if Items are added to the Collection,
        // (after 'isInstantiated = true')
        // the Groups which contain these added Items get rebuilt.
        // for (const key in this.groups) this.groups[key].rebuild();
    }
    /**
     * Updates the key/name identifier of the Collection.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/properties#key)
     *
     * @public
     * @param value - New key/name identifier.
     */
    set key(value) {
        this.setKey(value);
    }
    /**
     * Returns the key/name identifier of the Collection.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/properties#key)
     *
     * @public
     */
    get key() {
        return this._key;
    }
    /**
     * Updates the key/name identifier of the Collection.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#setkey)
     *
     * @public
     * @param value - New key/name identifier.
     */
    setKey(value) {
        const oldKey = this._key;
        // Update Collection key
        this._key = value;
        // Update key in Persistent (only if oldKey is equal to persistentKey
        // because otherwise the persistentKey is detached from the Collection key
        // -> not managed by Collection anymore)
        if (value != null && this.persistent?._key === oldKey)
            this.persistent?.setKey(value);
        return this;
    }
    /**
     * Creates a new Group without associating it to the Collection.
     *
     * This way of creating a Group is intended for use in the Collection configuration object,
     * where the `constructor()` takes care of the binding.
     *
     * After a successful initiation of the Collection we recommend using `createGroup()`,
     * because it automatically connects the Group to the Collection.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#group)
     *
     * @public
     * @param initialItems - Key/Name identifiers of the Items to be clustered by the Group.
     * @param config - Configuration object
     */
    Group(initialItems, config = {}) {
        if (this.isInstantiated) {
            const key = config.key ?? utils.generateId();
            LogCodeManager.log('1B:02:00');
            return this.createGroup(key, initialItems);
        }
        return new Group(this, initialItems, config);
    }
    /**
     * Creates a new Selector without associating it to the Collection.
     *
     * This way of creating a Selector is intended for use in the Collection configuration object,
     * where the `constructor()` takes care of the binding.
     *
     * After a successful initiation of the Collection we recommend using `createSelector()`,
     * because it automatically connects the Group to the Collection.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#selector)
     *
     * @public
     * @param initialKey - Key/Name identifier of the Item to be represented by the Selector.
     * @param config - Configuration object
     */
    Selector(initialKey, config = {}) {
        if (this.isInstantiated) {
            const key = config.key ?? utils.generateId();
            LogCodeManager.log('1B:02:01');
            return this.createSelector(key, initialKey);
        }
        return new Selector(this, initialKey, config);
    }
    /**
     * Sets up the specified Groups or Group keys
     * and assigns them to the Collection if they are valid.
     *
     * It also instantiates and assigns the default Group to the Collection.
     * The default Group reflects the default pattern of the Collection.
     *
     * @internal
     * @param groups - Entire Groups or Group keys to be set up.
     */
    initGroups(groups) {
        if (!groups)
            return;
        let groupsObject = {};
        // If groups is Array of Group keys/names, create the Groups based on these keys
        if (Array.isArray(groups)) {
            groups.forEach((groupKey) => {
                groupsObject[groupKey] = new Group(this, [], {
                    key: groupKey,
                });
            });
        }
        else
            groupsObject = groups;
        // Add default Group
        groupsObject[this.config.defaultGroupKey] = new Group(this, [], {
            key: this.config.defaultGroupKey,
        });
        // Assign missing key/name to Group based on the property key
        for (const key in groupsObject)
            if (groupsObject[key]._key == null)
                groupsObject[key].setKey(key);
        this.groups = groupsObject;
    }
    /**
     * Sets up the specified Selectors or Selector keys
     * and assigns them to the Collection if they are valid.
     *
     * @internal
     * @param selectors - Entire Selectors or Selector keys to be set up.
     */
    initSelectors(selectors) {
        if (!selectors)
            return;
        let selectorsObject = {};
        // If selectors is Array of Selector keys/names, create the Selectors based on these keys
        if (Array.isArray(selectors)) {
            selectors.forEach((selectorKey) => {
                selectorsObject[selectorKey] = new Selector(this, selectorKey, {
                    key: selectorKey,
                });
            });
        }
        else
            selectorsObject = selectors;
        // Assign missing key/name to Selector based on the property key
        for (const key in selectorsObject)
            if (selectorsObject[key]._key == null)
                selectorsObject[key].setKey(key);
        this.selectors = selectorsObject;
    }
    /**
     * Appends new data objects following the same pattern to the end of the Collection.
     *
     * Each collected `data object` requires a unique identifier at the primaryKey property (by default 'id')
     * to be correctly identified later.
     *
     * For example, if we collect some kind of user object,
     * it must contain such unique identifier at 'id'
     * to be added to the Collection.
     * ```
     * MY_COLLECTION.collect({id: '1', name: 'jeff'}); // valid
     * MY_COLLECTION.collect({name: 'frank'}); // invalid
     * ```
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#collect)
     *
     * @public
     * @param data - Data objects or entire Items to be added.
     * @param groupKeys - Group/s to which the specified data objects or Items are to be added.
     * @param config - Configuration object
     */
    collect(data, groupKeys, config = {}) {
        const _data = utils.normalizeArray(data);
        const _groupKeys = utils.normalizeArray(groupKeys);
        const defaultGroupKey = this.config.defaultGroupKey;
        const primaryKey = this.config.primaryKey;
        config = utils.defineConfig(config, {
            method: 'push',
            background: false,
            patch: false,
            select: false,
        });
        // Add default groupKey, since all Items are added to the default Group
        if (!_groupKeys.includes(defaultGroupKey))
            _groupKeys.push(defaultGroupKey);
        // Create not existing Groups
        _groupKeys.forEach((key) => this.groups[key] == null && this.createGroup(key));
        _data.forEach((data, index) => {
            let itemKey;
            let success = false;
            // Assign Data or Item to Collection
            if (data instanceof Item) {
                success = this.assignItem(data, {
                    background: config.background,
                });
                itemKey = data._key;
            }
            else {
                success = this.assignData(data, {
                    patch: config.patch,
                    background: config.background,
                });
                itemKey = data[primaryKey];
            }
            // Add itemKey to provided Groups and create corresponding Selector
            if (success) {
                _groupKeys.forEach((groupKey) => {
                    this.getGroup(groupKey)?.add(itemKey, {
                        method: config.method,
                        background: config.background,
                    });
                });
                if (config.select)
                    this.createSelector(itemKey, itemKey);
            }
            if (config.forEachItem)
                config.forEachItem(data, itemKey, success, index);
        });
        return this;
    }
    /**
     * Updates the Item `data object` with the specified `object with changes`, if the Item exists.
     * By default the `object with changes` is merged into the Item `data object` at top level.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#update)
     *
     * @public
     * @param itemKey - Key/Name identifier of the Item to be updated.
     * @param changes - Object with changes to be merged into the Item data object.
     * @param config - Configuration object
     */
    update(itemKey, changes, config = {}) {
        const item = this.getItem(itemKey, { notExisting: true });
        const primaryKey = this.config.primaryKey;
        config = utils.defineConfig(config, {
            patch: true,
            background: false,
        });
        // Check if the given conditions are suitable for a update action
        if (item == null) {
            LogCodeManager.log('1B:03:00', [itemKey, this._key]);
            return undefined;
        }
        if (!utils.isValidObject(changes)) {
            LogCodeManager.log('1B:03:01', [itemKey, this._key]);
            return undefined;
        }
        const oldItemKey = item._value[primaryKey];
        const newItemKey = changes[primaryKey] || oldItemKey;
        // Update itemKey if the new itemKey differs from the old one
        if (oldItemKey !== newItemKey)
            this.updateItemKey(oldItemKey, newItemKey, {
                background: config.background,
            });
        // Patch changes into Item data object
        if (config.patch) {
            // Delete primaryKey property from 'changes object' because if it has changed,
            // it is correctly updated in the above called 'updateItemKey()' method
            if (changes[primaryKey])
                delete changes[primaryKey];
            let patchConfig = typeof config.patch === 'object' ? config.patch : {};
            patchConfig = {
                addNewProperties: true,
                ...patchConfig,
            };
            item.patch(changes, {
                background: config.background,
                addNewProperties: patchConfig.addNewProperties,
            });
        }
        // Apply changes to Item data object
        else {
            // Ensure that the current Item identifier isn't different from the 'changes object' itemKey
            if (changes[this.config.primaryKey] !== itemKey) {
                changes[this.config.primaryKey] = itemKey;
                LogCodeManager.log('1B:02:02', [], changes);
            }
            item.set(changes, {
                background: config.background,
            });
        }
        return item;
    }
    /**
     * Creates a new Group and associates it to the Collection.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#createGroup)
     *
     * @public
     * @param groupKey - Unique identifier of the Group to be created.
     * @param initialItems - Key/Name identifiers of the Items to be clustered by the Group.
     */
    createGroup(groupKey, initialItems = []) {
        let group = this.getGroup(groupKey, { notExisting: true });
        if (!this.isInstantiated)
            LogCodeManager.log('1B:02:03');
        // Check if Group already exists
        if (group != null) {
            if (!group.isPlaceholder) {
                LogCodeManager.log('1B:03:02', [groupKey]);
                return group;
            }
            group.set(initialItems, { overwrite: true });
            return group;
        }
        // Create new Group
        group = new Group(this, initialItems, { key: groupKey });
        this.groups[groupKey] = group;
        return group;
    }
    /**
     * Returns a boolean indicating whether a Group with the specified `groupKey`
     * exists in the Collection or not.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#hasgroup)
     *
     * @public
     * @param groupKey - Key/Name identifier of the Group to be checked for existence.
     * @param config - Configuration object
     */
    hasGroup(groupKey, config = {}) {
        return !!this.getGroup(groupKey, config);
    }
    /**
     * Retrieves a single Group with the specified key/name identifier from the Collection.
     *
     * If the to retrieve Group doesn't exist, `undefined` is returned.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#getgroup)
     *
     * @public
     * @param groupKey - Key/Name identifier of the Group.
     * @param config - Configuration object
     */
    getGroup(groupKey, config = {}) {
        config = utils.defineConfig(config, {
            notExisting: false,
        });
        // Retrieve Group
        const group = groupKey ? this.groups[groupKey] : undefined;
        // Check if retrieved Group exists
        if (group == null || (!config.notExisting && !group.exists))
            return undefined;
        ComputedTracker.tracked(group.observers['value']);
        return group;
    }
    /**
     * Retrieves the default Group from the Collection.
     *
     * Every Collection should have a default Group,
     * which represents the default pattern of the Collection.
     *
     * If the default Group, for what ever reason, doesn't exist, `undefined` is returned.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#getdefaultgroup)
     *
     * @public
     */
    getDefaultGroup() {
        return this.getGroup(this.config.defaultGroupKey);
    }
    /**
     * Retrieves a single Group with the specified key/name identifier from the Collection.
     *
     * If the to retrieve Group doesn't exist, a reference Group is returned.
     * This has the advantage that Components that have the reference Group bound to themselves
     * are rerenderd when the original Group is created.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#getgroupwithreference)
     *
     * @public
     * @param groupKey - Key/Name identifier of the Group.
     */
    getGroupWithReference(groupKey) {
        let group = this.getGroup(groupKey, { notExisting: true });
        // Create dummy Group to hold reference
        if (group == null) {
            group = new Group(this, [], {
                key: groupKey,
                isPlaceholder: true,
            });
            this.groups[groupKey] = group;
        }
        ComputedTracker.tracked(group.observers['value']);
        return group;
    }
    /**
     * Removes a Group with the specified key/name identifier from the Collection,
     * if it exists in the Collection.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#removegroup)
     *
     * @public
     * @param groupKey - Key/Name identifier of the Group to be removed.
     */
    removeGroup(groupKey) {
        if (this.groups[groupKey] != null)
            delete this.groups[groupKey];
        return this;
    }
    /**
     * Returns the count of registered Groups in the Collection.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#getgroupcount)
     *
     * @public
     */
    getGroupCount() {
        let size = 0;
        Object.keys(this.groups).map(() => size++);
        return size;
    }
    /**
     * Creates a new Selector and associates it to the Collection.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#createSelector)
     *
     * @public
     * @param selectorKey - Unique identifier of the Selector to be created.
     * @param itemKey - Key/Name identifier of the Item to be represented by the Selector.
     */
    createSelector(selectorKey, itemKey) {
        let selector = this.getSelector(selectorKey, { notExisting: true });
        if (!this.isInstantiated)
            LogCodeManager.log('1B:02:04');
        // Check if Selector already exists
        if (selector != null) {
            if (!selector.isPlaceholder) {
                LogCodeManager.log('1B:03:03', [selectorKey]);
                return selector;
            }
            selector.select(itemKey, { overwrite: true });
            return selector;
        }
        // Create new Selector
        selector = new Selector(this, itemKey, {
            key: selectorKey,
        });
        this.selectors[selectorKey] = selector;
        return selector;
    }
    /**
     * Creates a new Selector and associates it to the Collection.
     *
     * The specified `itemKey` is used as the unique identifier key of the new Selector.
     * ```
     * MY_COLLECTION.select('1');
     * // is equivalent to
     * MY_COLLECTION.createSelector('1', '1');
     * ```
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#select)
     *
     * @public
     * @param itemKey - Key/Name identifier of the Item to be represented by the Selector
     * and used as unique identifier of the Selector.
     */
    select(itemKey) {
        return this.createSelector(itemKey, itemKey);
    }
    /**
     * Returns a boolean indicating whether a Selector with the specified `selectorKey`
     * exists in the Collection or not.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#hasselector)
     *
     * @public
     * @param selectorKey - Key/Name identifier of the Selector to be checked for existence.
     * @param config - Configuration object
     */
    hasSelector(selectorKey, config = {}) {
        return !!this.getSelector(selectorKey, config);
    }
    /**
     * Retrieves a single Selector with the specified key/name identifier from the Collection.
     *
     * If the to retrieve Selector doesn't exist, `undefined` is returned.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#getselector)
     *
     * @public
     * @param selectorKey - Key/Name identifier of the Selector.
     * @param config - Configuration object
     */
    getSelector(selectorKey, config = {}) {
        config = utils.defineConfig(config, {
            notExisting: false,
        });
        // Get Selector
        const selector = selectorKey ? this.selectors[selectorKey] : undefined;
        // Check if Selector exists
        if (selector == null || (!config.notExisting && !selector.exists))
            return undefined;
        ComputedTracker.tracked(selector.observers['value']);
        return selector;
    }
    /**
     * Retrieves a single Selector with the specified key/name identifier from the Collection.
     *
     * If the to retrieve Selector doesn't exist, a reference Selector is returned.
     * This has the advantage that Components that have the reference Selector bound to themselves
     * are rerenderd when the original Selector is created.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#getselectorwithreference)
     *
     * @public
     * @param selectorKey - Key/Name identifier of the Selector.
     */
    getSelectorWithReference(selectorKey) {
        let selector = this.getSelector(selectorKey, { notExisting: true });
        // Create dummy Selector to hold reference
        if (selector == null) {
            selector = new Selector(this, null, {
                key: selectorKey,
                isPlaceholder: true,
            });
            this.selectors[selectorKey] = selector;
        }
        ComputedTracker.tracked(selector.observers['value']);
        return selector;
    }
    /**
     * Removes a Selector with the specified key/name identifier from the Collection,
     * if it exists in the Collection.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#removeselector)
     *
     * @public
     * @param selectorKey - Key/Name identifier of the Selector to be removed.
     */
    removeSelector(selectorKey) {
        if (this.selectors[selectorKey] != null) {
            this.selectors[selectorKey].unselect();
            delete this.selectors[selectorKey];
        }
        return this;
    }
    /**
     * Returns the count of registered Selectors in the Collection.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#getselectorcount)
     *
     * @public
     */
    getSelectorCount() {
        let size = 0;
        Object.keys(this.selectors).map(() => size++);
        return size;
    }
    /**
     * Returns a boolean indicating whether a Item with the specified `itemKey`
     * exists in the Collection or not.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#hasitem)
     *
     * @public
     * @param itemKey - Key/Name identifier of the Item.
     * @param config - Configuration object
     */
    hasItem(itemKey, config = {}) {
        return !!this.getItem(itemKey, config);
    }
    /**
     * Retrieves a single Item with the specified key/name identifier from the Collection.
     *
     * If the to retrieve Item doesn't exist, `undefined` is returned.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#getitem)
     *
     * @public
     * @param itemKey - Key/Name identifier of the Item.
     * @param config - Configuration object
     */
    getItem(itemKey, config = {}) {
        config = utils.defineConfig(config, {
            notExisting: false,
        });
        // Get Item
        const item = itemKey != null ? this.data[itemKey] : undefined;
        // Check if Item exists
        if (item == null || (!config.notExisting && !item.exists))
            return undefined;
        ComputedTracker.tracked(item.observers['value']);
        return item;
    }
    /**
     * Retrieves a single Item with the specified key/name identifier from the Collection.
     *
     * If the to retrieve Item doesn't exist, a reference Item is returned.
     * This has the advantage that Components that have the reference Item bound to themselves
     * are rerenderd when the original Item is created.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#getitemwithreference)
     *
     * @public
     * @param itemKey - Key/Name identifier of the Item.
     */
    getItemWithReference(itemKey) {
        let item = this.getItem(itemKey, { notExisting: true });
        // Create dummy Item to hold reference
        if (item == null)
            item = this.createPlaceholderItem(itemKey, true);
        ComputedTracker.tracked(item.observers['value']);
        return item;
    }
    /**
     * Creates a placeholder Item
     * that can be used to hold a reference to a not existing Item.
     *
     * @internal
     * @param itemKey - Unique identifier of the to create placeholder Item.
     * @param addToCollection - Whether to add the Item to be created to the Collection.
     */
    createPlaceholderItem(itemKey, addToCollection = false) {
        // Create placeholder Item
        const item = new Item(this, {
            [this.config.primaryKey]: itemKey,
            dummy: 'item',
        }, { isPlaceholder: true });
        // Add placeholder Item to Collection
        if (addToCollection &&
            !Object.prototype.hasOwnProperty.call(this.data, itemKey))
            this.data[itemKey] = item;
        ComputedTracker.tracked(item.observers['value']);
        return item;
    }
    /**
     * Retrieves the value (data object) of a single Item
     * with the specified key/name identifier from the Collection.
     *
     * If the to retrieve Item containing the value doesn't exist, `undefined` is returned.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#getitemvalue)
     *
     * @public
     * @param itemKey - Key/Name identifier of the Item.
     * @param config - Configuration object
     */
    getItemValue(itemKey, config = {}) {
        const item = this.getItem(itemKey, config);
        if (item == null)
            return undefined;
        return item.value;
    }
    /**
     * Retrieves all Items from the Collection.
     * ```
     * MY_COLLECTION.getAllItems();
     * // is equivalent to
     * MY_COLLECTION.getDefaultGroup().items;
     * ```
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#getallitems)
     *
     * @public
     * @param config - Configuration object
     */
    getAllItems(config = {}) {
        config = utils.defineConfig(config, {
            notExisting: false,
        });
        const defaultGroup = this.getDefaultGroup();
        let items = [];
        // If config.notExisting transform the data object into array since it contains all Items,
        // otherwise return the default Group Items
        if (config.notExisting) {
            for (const key in this.data)
                items.push(this.data[key]);
        }
        else {
            // Why default Group Items and not all '.exists === true' Items?
            // Because the default Group keeps track of all existing Items.
            // It also does control the Collection output in binding methods like 'useAgile()'
            // and therefore should do it here too.
            items = defaultGroup?.getItems() || [];
        }
        return items;
    }
    /**
     * Retrieves the values (data objects) of all Items from the Collection.
     * ```
     * MY_COLLECTION.getAllItemValues();
     * // is equivalent to
     * MY_COLLECTION.getDefaultGroup().output;
     * ```
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#getallitemvalues)
     *
     * @public
     * @param config - Configuration object
     */
    getAllItemValues(config = {}) {
        const items = this.getAllItems(config);
        return items.map((item) => item.value);
    }
    persist(keyOrConfig = {}, config = {}) {
        let _config;
        let key;
        if (utils.isValidObject(keyOrConfig)) {
            _config = keyOrConfig;
            key = this._key;
        }
        else {
            _config = config || {};
            key = keyOrConfig;
        }
        _config = utils.defineConfig(_config, {
            loadValue: true,
            storageKeys: [],
            defaultStorageKey: null,
        });
        // Check if Collection is already persisted
        if (this.persistent != null && this.isPersisted)
            return this;
        // Create Persistent (-> persist value)
        this.persistent = new CollectionPersistent(this, {
            loadValue: _config.loadValue,
            storageKeys: _config.storageKeys,
            key: key,
            defaultStorageKey: _config.defaultStorageKey,
        });
        return this;
    }
    /**
     * Fires immediately after the persisted `value`
     * is loaded into the Collection from a corresponding external Storage.
     *
     * Registering such callback function makes only sense
     * when the Collection is [persisted](https://agile-ts.org/docs/core/collection/methods/#persist) in an external Storage.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#onload)
     *
     * @public
     * @param callback - A function to be executed after the externally persisted `value` was loaded into the Collection.
     */
    onLoad(callback) {
        if (!this.persistent)
            return this;
        if (!utils.isFunction(callback)) {
            LogCodeManager.log('00:03:01', ['OnLoad Callback', 'function']);
            return this;
        }
        // Register specified callback
        this.persistent.onLoad = callback;
        // If Collection is already persisted ('isPersisted') fire specified callback immediately
        if (this.isPersisted)
            callback(true);
        return this;
    }
    /**
     * Removes all Items from the Collection
     * and resets all Groups and Selectors of the Collection.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#reset)
     *
     * @public
     */
    reset() {
        // Reset data
        this.data = {};
        this.size = 0;
        // Reset Groups
        for (const key in this.groups)
            this.getGroup(key)?.reset();
        // Reset Selectors
        for (const key in this.selectors)
            this.getSelector(key)?.reset();
        return this;
    }
    /**
     * Puts `itemKeys/s` into Group/s.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#put)
     *
     * @public
     * @param itemKeys - `itemKey/s` to be put into the specified Group/s.
     * @param groupKeys - Key/Name Identifier/s of the Group/s the specified `itemKey/s` are to put in.
     * @param config - Configuration object
     */
    put(itemKeys, groupKeys, config = {}) {
        const _itemKeys = utils.normalizeArray(itemKeys);
        const _groupKeys = utils.normalizeArray(groupKeys);
        // Assign itemKeys to Groups
        _groupKeys.forEach((groupKey) => {
            this.getGroup(groupKey)?.add(_itemKeys, config);
        });
        return this;
    }
    /**
     * Moves specified `itemKey/s` from one Group to another Group.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#move)
     *
     * @public
     * @param itemKeys - `itemKey/s` to be moved.
     * @param oldGroupKey - Key/Name Identifier of the Group the `itemKey/s` are moved from.
     * @param newGroupKey - Key/Name Identifier of the Group the `itemKey/s` are moved in.
     * @param config - Configuration object
     */
    move(itemKeys, oldGroupKey, newGroupKey, config = {}) {
        const _itemKeys = utils.normalizeArray(itemKeys);
        // Remove itemKeys from old Group
        this.getGroup(oldGroupKey)?.remove(_itemKeys, utils.removeProperties(config, ['method', 'overwrite']));
        // Assign itemKeys to new Group
        this.getGroup(newGroupKey)?.add(_itemKeys, config);
        return this;
    }
    /**
     * Updates the key/name identifier of the Item
     * and returns a boolean indicating
     * whether the Item identifier was updated successfully.
     *
     * @internal
     * @param oldItemKey - Old key/name Item identifier.
     * @param newItemKey - New key/name Item identifier.
     * @param config - Configuration object
     */
    updateItemKey(oldItemKey, newItemKey, config = {}) {
        const item = this.getItem(oldItemKey, { notExisting: true });
        config = utils.defineConfig(config, {
            background: false,
        });
        if (item == null || oldItemKey === newItemKey)
            return false;
        // Check if Item with newItemKey already exists
        if (this.hasItem(newItemKey)) {
            LogCodeManager.log('1B:03:04', [oldItemKey, newItemKey, this._key]);
            return false;
        }
        // Update itemKey in data object
        delete this.data[oldItemKey];
        this.data[newItemKey] = item;
        // Update key/name of the Item
        item.setKey(newItemKey, {
            background: config.background,
        });
        // Update Persistent key of the Item if it follows the Item Storage Key pattern
        // and therefore differs from the actual Item key
        // (-> isn't automatically updated when the Item key is updated)
        if (item.persistent != null &&
            item.persistent._key ===
                CollectionPersistent.getItemStorageKey(oldItemKey, this._key))
            item.persistent?.setKey(CollectionPersistent.getItemStorageKey(newItemKey, this._key));
        // Update itemKey in Groups
        for (const groupKey in this.groups) {
            const group = this.getGroup(groupKey, { notExisting: true });
            if (group == null || !group.has(oldItemKey))
                continue;
            group.replace(oldItemKey, newItemKey, { background: config.background });
        }
        // Update itemKey in Selectors
        for (const selectorKey in this.selectors) {
            const selector = this.getSelector(selectorKey, { notExisting: true });
            if (selector == null)
                continue;
            // Reselect Item in Selector that has selected the newItemKey.
            // Necessary because potential reference placeholder Item got overwritten
            // with the new (renamed) Item
            // -> has to find the new Item at selected itemKey
            //    since the placeholder Item got overwritten
            if (selector.hasSelected(newItemKey, false)) {
                selector.reselect({
                    force: true,
                    background: config.background,
                });
            }
            // Select newItemKey in Selector that has selected the oldItemKey
            if (selector.hasSelected(oldItemKey, false))
                selector.select(newItemKey, {
                    background: config.background,
                });
        }
        return true;
    }
    /**
     * Returns all key/name identifiers of the Group/s containing the specified `itemKey`.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#getgroupkeysthathaveitemkey)
     *
     * @public
     * @param itemKey - `itemKey` to be contained in Group/s.
     */
    getGroupKeysThatHaveItemKey(itemKey) {
        const groupKeys = [];
        for (const groupKey in this.groups) {
            const group = this.groups[groupKey];
            if (group?.has(itemKey))
                groupKeys.push(groupKey);
        }
        return groupKeys;
    }
    /**
     * Removes Item/s from:
     *
     * - `.everywhere()`:
     *    Removes Item/s from the entire Collection and all its Groups and Selectors (i.e. from everywhere)
     *    ```
     *    MY_COLLECTION.remove('1').everywhere();
     *    // is equivalent to
     *    MY_COLLECTION.removeItems('1');
     *    ```
     * - `.fromGroups()`:
     *   Removes Item/s only from specified Groups.
     *   ```
     *    MY_COLLECTION.remove('1').fromGroups(['1', '2']);
     *    // is equivalent to
     *    MY_COLLECTION.removeFromGroups('1', ['1', '2']);
     *    ```
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#remove)
     *
     * @public
     * @param itemKeys - Item/s with identifier/s to be removed.
     */
    remove(itemKeys) {
        return {
            fromGroups: (groups) => this.removeFromGroups(itemKeys, groups),
            everywhere: (config) => this.removeItems(itemKeys, config || {}),
        };
    }
    /**
     * Remove Item/s from specified Group/s.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#removefromgroups)
     *
     * @public
     * @param itemKeys - Key/Name Identifier/s of the Item/s to be removed from the Group/s.
     * @param groupKeys - Key/Name Identifier/s of the Group/s the Item/s are to remove from.
     */
    removeFromGroups(itemKeys, groupKeys) {
        const _itemKeys = utils.normalizeArray(itemKeys);
        const _groupKeys = utils.normalizeArray(groupKeys);
        _itemKeys.forEach((itemKey) => {
            let removedFromGroupsCount = 0;
            // Remove itemKey from the Groups
            _groupKeys.forEach((groupKey) => {
                const group = this.getGroup(groupKey, { notExisting: true });
                if (!group?.has(itemKey))
                    return;
                group.remove(itemKey);
                removedFromGroupsCount++;
            });
            // If the Item was removed from each Group representing the Item,
            // remove it completely
            if (removedFromGroupsCount >=
                this.getGroupKeysThatHaveItemKey(itemKey).length)
                this.removeItems(itemKey);
        });
        return this;
    }
    /**
     * Removes Item/s from the entire Collection and all the Collection's Groups and Selectors.
     *
     * [Learn more..](https://agile-ts.org/docs/core/collection/methods/#removeitems)
     *
     * @public
     * @param itemKeys - Key/Name identifier/s of the Item/s to be removed from the entire Collection.
     * @param config - Configuration object
     */
    removeItems(itemKeys, config = {}) {
        config = utils.defineConfig(config, {
            notExisting: false,
            removeSelector: false,
        });
        const _itemKeys = utils.normalizeArray(itemKeys);
        _itemKeys.forEach((itemKey) => {
            const item = this.getItem(itemKey, { notExisting: config.notExisting });
            if (item == null)
                return;
            const wasPlaceholder = item.isPlaceholder;
            // Remove Item from the Groups
            for (const groupKey in this.groups) {
                const group = this.getGroup(groupKey, { notExisting: true });
                if (group?.has(itemKey))
                    group?.remove(itemKey);
            }
            // Remove Item from Storage
            item.persistent?.removePersistedValue();
            // Remove Item from Collection
            delete this.data[itemKey];
            // Reselect or remove Selectors which have represented the removed Item
            for (const selectorKey in this.selectors) {
                const selector = this.getSelector(selectorKey, { notExisting: true });
                if (selector != null && selector.hasSelected(itemKey, false)) {
                    if (config.removeSelector) {
                        // Remove Selector
                        this.removeSelector(selector._key ?? 'unknown');
                    }
                    else {
                        // Reselect Item in Selector
                        // in order to create a new dummyItem
                        // to hold a reference to the now not existing Item
                        selector.reselect({ force: true });
                    }
                }
            }
            if (!wasPlaceholder)
                this.size--;
        });
        return this;
    }
    /**
     * Assigns the provided `data` object to an already existing Item
     * with specified key/name identifier found in the `data` object.
     * If the Item doesn't exist yet, a new Item with the `data` object as value
     * is created and assigned to the Collection.
     *
     * Returns a boolean indicating
     * whether the `data` object was assigned/updated successfully.
     *
     * @internal
     * @param data - Data object
     * @param config - Configuration object
     */
    assignData(data, config = {}) {
        config = utils.defineConfig(config, {
            patch: false,
            background: false,
        });
        const _data = utils.copy(data); // Copy data object to get rid of reference
        const primaryKey = this.config.primaryKey;
        if (!utils.isValidObject(_data)) {
            LogCodeManager.log('1B:03:05', [this._key]);
            return false;
        }
        // Check if data object contains valid itemKey,
        // otherwise add random itemKey to Item
        if (!Object.prototype.hasOwnProperty.call(_data, primaryKey)) {
            LogCodeManager.log('1B:02:05', [this._key, primaryKey]);
            _data[primaryKey] = utils.generateId();
        }
        const itemKey = _data[primaryKey];
        const item = this.getItem(itemKey, { notExisting: true });
        const wasPlaceholder = item?.isPlaceholder || false;
        // Create new Item or update existing Item
        if (item != null) {
            if (config.patch) {
                item.patch(_data, { background: config.background });
            }
            else {
                item.set(_data, { background: config.background });
            }
        }
        else {
            this.assignItem(new Item(this, _data), {
                background: config.background,
            });
        }
        // Increase size of Collection if Item was previously a placeholder
        // (-> hasn't officially existed in Collection before)
        if (wasPlaceholder)
            this.size++;
        return true;
    }
    /**
     * Assigns the specified Item to the Collection
     * at the key/name identifier of the Item.
     *
     * And returns a boolean indicating
     * whether the Item was assigned successfully.
     *
     * @internal
     * @param item - Item to be added.
     * @param config - Configuration object
     */
    assignItem(item, config = {}) {
        config = utils.defineConfig(config, {
            overwrite: false,
            background: false,
            rebuildGroups: true,
        });
        const primaryKey = this.config.primaryKey;
        let itemKey = item._value[primaryKey];
        let increaseCollectionSize = true;
        // Check if Item has valid itemKey,
        // otherwise add random itemKey to Item
        if (!Object.prototype.hasOwnProperty.call(item._value, primaryKey)) {
            LogCodeManager.log('1B:02:05', [this._key, primaryKey]);
            itemKey = utils.generateId();
            item.patch({ [this.config.primaryKey]: itemKey }, { background: config.background });
            item._key = itemKey;
        }
        // Check if Item belongs to this Collection
        if (item.collection() !== this) {
            LogCodeManager.log('1B:03:06', [this._key, item.collection()._key]);
            return false;
        }
        // Check if Item already exists
        if (this.getItem(itemKey, { notExisting: true }) != null) {
            if (!config.overwrite) {
                this.assignData(item._value);
                return true;
            }
            else
                increaseCollectionSize = false;
        }
        // Assign/add Item to Collection
        this.data[itemKey] = item;
        // Rebuild Groups that include itemKey
        // after adding Item with itemKey to the Collection
        // (because otherwise it can't find the Item as it isn't added yet)
        if (config.rebuildGroups)
            this.rebuildGroupsThatIncludeItemKey(itemKey, {
                background: config.background,
            });
        if (increaseCollectionSize)
            this.size++;
        return true;
    }
    /**
     * Rebuilds all Groups that contain the specified `itemKey`.
     *
     * @internal
     * @itemKey - `itemKey` Groups must contain to be rebuilt.
     * @config - Configuration object
     */
    rebuildGroupsThatIncludeItemKey(itemKey, config = {}) {
        // Rebuild Groups that include itemKey
        for (const groupKey of Object.keys(this.groups)) {
            const group = this.getGroup(groupKey);
            if (group != null && group.has(itemKey)) {
                const index = group._preciseItemKeys.findIndex((ik) => itemKey === ik);
                // Update Group output at index
                if (index !== -1) {
                    group.rebuild([
                        {
                            key: itemKey,
                            index: index,
                            method: exports.TrackedChangeMethod.UPDATE,
                        },
                    ], config);
                }
                // Add Item to the Group output if it isn't yet represented there to be updated
                else {
                    const indexOfBeforeItemKey = group.nextStateValue.findIndex((ik) => itemKey === ik) - 1;
                    group.rebuild([
                        {
                            key: itemKey,
                            index: indexOfBeforeItemKey >= 0
                                ? group._preciseItemKeys.findIndex((ik) => group.nextStateValue[indexOfBeforeItemKey] === ik) + 1
                                : 0,
                            method: exports.TrackedChangeMethod.ADD,
                        },
                    ], config);
                }
            }
        }
    }
}

/**
 * Returns a newly created Collection.
 *
 * A Collection manages a reactive set of Information
 * that we need to remember globally at a later point in time.
 * While providing a toolkit to use and mutate this set of Information.
 *
 * It is designed for arrays of data objects following the same pattern.
 *
 * Each of these data object must have a unique `primaryKey` to be correctly identified later.
 *
 * You can create as many global Collections as you need.
 *
 * [Learn more..](https://agile-ts.org/docs/core/agile-instance/methods#createcollection)
 *
 * @public
 * @param config - Configuration object
 * @param agileInstance - Instance of Agile the Collection belongs to.
 */
function createCollection(config, agileInstance = exports.sharedAgileInstance) {
    return new Collection(agileInstance, config);
}

var Agile = Agile$1;

exports.Agile = Agile$1;
exports.CallbackSubscriptionContainer = CallbackSubscriptionContainer;
exports.Collection = Collection;
exports.CollectionPersistent = CollectionPersistent;
exports.ComponentSubscriptionContainer = ComponentSubscriptionContainer;
exports.Computed = Computed;
exports.ComputedTracker = ComputedTracker;
exports.EnhancedState = EnhancedState;
exports.Group = Group;
exports.GroupObserver = GroupObserver;
exports.Integration = Integration;
exports.Integrations = Integrations;
exports.Item = Item;
exports.LogCodeManager = LogCodeManager;
exports.Observer = Observer;
exports.Persistent = Persistent;
exports.Runtime = Runtime;
exports.RuntimeJob = RuntimeJob;
exports.Selector = Selector;
exports.State = State;
exports.StateObserver = StateObserver;
exports.StatePersistent = StatePersistent;
exports.StateRuntimeJob = StateRuntimeJob;
exports.Storage = Storage;
exports.Storages = Storages;
exports.SubController = SubController;
exports.SubscriptionContainer = SubscriptionContainer;
exports.assignAdditionalLogs = assignAdditionalLogs;
exports.assignSharedAgileInstance = assignSharedAgileInstance;
exports.assignSharedStorageManager = assignSharedStorageManager;
exports.createCollection = createCollection;
exports.createComputed = createComputed;
exports.createEnhancedState = createEnhancedState;
exports.createLightState = createLightState;
exports.createState = createState;
exports.createStorage = createStorage;
exports.createStorageManager = createStorageManager;
exports['default'] = Agile;
exports.extractObservers = extractObservers;
exports.extractRelevantObservers = extractRelevantObservers;
exports.getAgileInstance = getAgileInstance;
exports.getSharedStorageManager = getSharedStorageManager;
exports.globalBind = globalBind;
exports.optionalRequire = optionalRequire;
exports.runsOnServer = runsOnServer;
exports.shared = exports.sharedAgileInstance;
Object.keys(utils).forEach(function (k) {
    if (k !== 'default' && !exports.hasOwnProperty(k)) Object.defineProperty(exports, k, {
        enumerable: true,
        get: function () {
            return utils[k];
        }
    });
});
